from propan import PropanApp, RabbitBroker
from propan.annotations import Logger
from propan.brokers.rabbit import RabbitQueue

broker = RabbitBroker(consumers=10)
app = PropanApp(broker)

queue = RabbitQueue(
    name="test",
    durable=True,
    arguments={
        "x-queue-type": "stream",
    },
)


@broker.handle(queue, consume_arguments={"x-stream-offset": "first"})
async def handle(msg, logger: Logger):
    logger.info(msg)


@app.after_startup
async def test():
    await broker.publish("Hi!", queue)



from typing import Dict, List, Optional

from pydantic import BaseModel, Field

from propan.asyncapi.channels import AsyncAPIChannel
from propan.asyncapi.info import AsyncAPIInfo
from propan.asyncapi.message import AsyncAPIMessage
from propan.asyncapi.servers import AsyncAPIServer
from propan.asyncapi.utils import AsyncAPIExternalDocs, AsyncAPITag
from propan.brokers.constants import ContentTypes
from propan.types import AnyDict

ASYNC_API_VERSION = "2.6.0"


class AsyncAPIComponents(BaseModel):
    # TODO
    # servers
    # serverVariavles
    # channels
    messages: Optional[Dict[str, AsyncAPIMessage]] = None
    schemas: Optional[Dict[str, AnyDict]] = None

    # securitySchemes
    # parameters
    # correlationIds
    # operationTraits
    # messageTraits
    # serverBindings
    # channelBindings
    # operationBindings
    # messageBindings


class AsyncAPISchema(BaseModel):
    asyncapi: str = ASYNC_API_VERSION
    default_content_type: str = Field(
        default=ContentTypes.json.value,
        alias="defaultContentType",
    )
    info: AsyncAPIInfo
    servers: Optional[Dict[str, AsyncAPIServer]] = None
    channels: Dict[str, AsyncAPIChannel]
    tags: Optional[List[AsyncAPITag]] = None
    external_docs: Optional[AsyncAPIExternalDocs] = Field(
        default=None,
        alias="externalDocs",
    )

    # TODO:
    # id
    components: Optional[AsyncAPIComponents] = None



import asyncio
import inspect
from functools import wraps
from itertools import dropwhile
from typing import Any, Callable, Coroutine, Optional, Sequence, Union

from fastapi import params
from fastapi.dependencies.models import Dependant
from fastapi.dependencies.utils import (
    get_dependant,
    get_parameterless_sub_dependant,
    solve_dependencies,
)
from fastapi.routing import run_endpoint_function
from starlette.requests import Request
from starlette.routing import BaseRoute

from propan._compat import raise_fastapi_validation_error
from propan.brokers._model.broker_usecase import (
    BrokerAsyncUsecase,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers._model.schemas import PropanMessage as NativeMessage
from propan.brokers._model.schemas import Queue
from propan.types import AnyDict


class PropanRoute(BaseRoute):
    def __init__(
        self,
        path: Union[Queue, str],
        *extra: Union[Queue, str],
        endpoint: HandlerCallable[T_HandlerReturn],
        broker: BrokerAsyncUsecase[Any, Any],
        dependencies: Sequence[params.Depends] = (),
        dependency_overrides_provider: Optional[Any] = None,
        **handle_kwargs: AnyDict,
    ) -> None:
        self.path = path
        self.broker = broker

        path_name = (path if isinstance(path, str) else path.name) or ""
        dependant = get_dependant(
            path=path_name,
            call=endpoint,
        )
        for depends in dependencies[::-1]:
            dependant.dependencies.insert(
                0,
                get_parameterless_sub_dependant(depends=depends, path=path_name),
            )

        handler = wraps(endpoint)(
            PropanMessage.get_session(
                dependant,
                dependency_overrides_provider,
            )
        )

        self.dependant = dependant

        self.handler = broker.handle(
            path,
            *extra,
            _raw=True,
            _get_dependant=lambda call: dependant,
            **handle_kwargs,  # type: ignore
        )(handler)


class PropanMessage(Request):
    scope: AnyDict
    _cookies: AnyDict
    _headers: AnyDict  # type: ignore
    _body: AnyDict  # type: ignore
    _query_params: AnyDict  # type: ignore

    def __init__(
        self,
        body: Optional[AnyDict] = None,
        headers: Optional[AnyDict] = None,
    ):
        self.scope = {}
        self._cookies = {}
        self._headers = headers or {}
        self._body = body or {}
        self._query_params = self._body

    @classmethod
    def get_session(
        cls,
        dependant: Dependant,
        dependency_overrides_provider: Optional[Any] = None,
    ) -> Callable[[NativeMessage[Any]], Any]:
        assert dependant.call
        func = get_app(dependant, dependency_overrides_provider)

        dependencies_names = tuple(i.name for i in dependant.dependencies)

        first_arg = next(
            dropwhile(
                lambda i: i in dependencies_names,
                inspect.signature(dependant.call).parameters,
            ),
            None,
        )

        async def app(message: NativeMessage[Any]) -> Any:
            body = message.decoded_body
            if first_arg is not None:
                if not isinstance(body, dict):  # pragma: no branch
                    fastapi_body: AnyDict = {first_arg: body}
                else:
                    fastapi_body = body

                session = cls(fastapi_body, message.headers)
            else:
                session = cls()
            return await func(session)

        return app


def get_app(
    dependant: Dependant,
    dependency_overrides_provider: Optional[Any] = None,
) -> Callable[[PropanMessage], Coroutine[Any, Any, Any]]:
    async def app(request: PropanMessage) -> Any:
        solved_result = await solve_dependencies(
            request=request,
            body=request._body,
            dependant=dependant,
            dependency_overrides_provider=dependency_overrides_provider,
        )

        values, errors, _, _2, _3 = solved_result
        if errors:
            raise_fastapi_validation_error(errors, request._body)

        return await run_endpoint_function(
            dependant=dependant,
            values=values,
            is_coroutine=asyncio.iscoroutinefunction(dependant.call),
        )

    return app



from types import TracebackType
from typing import Generic, Optional, Type

from typing_extensions import TypeVar

from propan.brokers._model.schemas import PropanMessage

MsgType = TypeVar("MsgType")


class BaseMiddleware(Generic[MsgType]):
    message: PropanMessage[MsgType]

    def __init__(self, msg: PropanMessage[MsgType]) -> None:
        self.message = msg

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_val: Optional[BaseException] = None,
        exec_tb: Optional[TracebackType] = None,
    ):
        pass

    async def __aenter__(self):
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_val: Optional[BaseException] = None,
        exec_tb: Optional[TracebackType] = None,
    ):
        pass



import asyncio
import logging
from functools import wraps
from secrets import token_hex
from types import TracebackType
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

import anyio
import nats
from fast_depends.dependencies import Depends
from nats.aio.client import Callback, Client, ErrorCallback
from nats.aio.msg import Msg
from nats.aio.subscription import (
    DEFAULT_SUB_PENDING_BYTES_LIMIT,
    DEFAULT_SUB_PENDING_MSGS_LIMIT,
)
from typing_extensions import TypeAlias

from propan.brokers._model.broker_usecase import (
    BrokerAsyncUsecase,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers._model.schemas import PropanMessage
from propan.brokers.exceptions import WRONG_PUBLISH_ARGS
from propan.brokers.nats.schemas import Handler
from propan.brokers.push_back_watcher import BaseWatcher
from propan.types import AnyDict, DecodedMessage, SendableMessage
from propan.utils import context

T = TypeVar("T")
NatsMessage: TypeAlias = PropanMessage[Msg]


class NatsBroker(BrokerAsyncUsecase[Msg, Client]):
    handlers: List[Handler]

    _max_queue_len: int
    _max_subject_len: int
    __is_connected: bool

    def __init__(
        self,
        servers: Union[str, List[str]] = ["nats://localhost:4222"],  # noqa: B006
        *,
        log_fmt: Optional[str] = None,
        protocol: str = "nats",
        **kwargs: AnyDict,
    ) -> None:
        super().__init__(
            servers, log_fmt=log_fmt, url_=servers, protocol=protocol, **kwargs
        )

        self._connection = None

        self._max_queue_len = 0
        self._max_subject_len = 4
        self.__is_connected = True

    async def _connect(
        self,
        *,
        url: Optional[str] = None,
        error_cb: Optional[ErrorCallback] = None,
        reconnected_cb: Optional[Callback] = None,
        **kwargs: Any,
    ) -> Client:
        if url is not None:
            kwargs["servers"] = kwargs.pop("servers", []) + [url]
        return await nats.connect(
            error_cb=self.log_connection_broken(error_cb),
            reconnected_cb=self.log_reconnected(reconnected_cb),
            **kwargs,
        )

    def handle(
        self,
        subject: str,
        queue: str = "",
        *,
        dependencies: Sequence[Depends] = (),
        pending_msgs_limit: int = DEFAULT_SUB_PENDING_MSGS_LIMIT,
        pending_bytes_limit: int = DEFAULT_SUB_PENDING_BYTES_LIMIT,
        description: str = "",
        **original_kwargs: AnyDict,
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[Msg, bool], Awaitable[T_HandlerReturn]],
    ]:
        super().handle()

        self._max_subject_len = max((self._max_subject_len, len(subject)))
        self._max_queue_len = max((self._max_queue_len, len(queue)))

        def wrapper(
            func: HandlerCallable[T_HandlerReturn],
        ) -> Callable[[Msg, bool], Awaitable[T_HandlerReturn]]:
            func, dependant = self._wrap_handler(
                func,
                queue=queue,
                subject=subject,
                extra_dependencies=dependencies,
                **original_kwargs,
            )
            handler = Handler(
                callback=func,
                subject=subject,
                queue=queue,
                _description=description,
                dependant=dependant,
                extra_args={
                    "pending_msgs_limit": pending_msgs_limit,
                    "pending_bytes_limit": pending_bytes_limit,
                },
            )
            self.handlers.append(handler)

            return func

        return wrapper

    async def start(self) -> None:
        context.set_local(
            "log_context",
            self._get_log_context(None, ""),
        )

        await self._start()

        for handler in self.handlers:
            func = handler.callback

            c = self._get_log_context(None, handler.subject, handler.queue)
            self._log(f"`{func.__name__}` waiting for messages", extra=c)

            sub = await self._connection.subscribe(
                subject=handler.subject,
                queue=handler.queue,
                cb=func,
                **handler.extra_args,
            )
            handler.subscription = sub

    async def _start(self):
        await super().start()

    async def publish(
        self,
        message: SendableMessage,
        subject: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        reply_to: str = "",
        callback: bool = False,
        callback_timeout: Optional[float] = 30.0,
        raise_timeout: bool = False,
    ) -> Optional[DecodedMessage]:
        if self._connection is None:
            raise ValueError("NatsConnection not started yet")

        msg, content_type = self._encode_message(message)

        client = self._connection

        if callback is True:
            if reply_to:
                raise WRONG_PUBLISH_ARGS

            token = client._nuid.next()
            token.extend(token_hex(2).encode())
            reply_to = token.decode()

        if reply_to:
            future: asyncio.Future[Msg] = asyncio.Future()
            sub = await client.subscribe(reply_to, future=future, max_msgs=1)
            await sub.unsubscribe(limit=1)

            kwargs = {"reply": reply_to}

        else:
            kwargs = {}

        await client.publish(
            subject=subject,
            payload=msg,
            headers={
                **(headers or {}),
                "content-type": content_type or "",
            },
            **kwargs,
        )

        if reply_to:
            if raise_timeout:
                scope = anyio.fail_after
            else:
                scope = anyio.move_on_after

            msg: Any = None
            with scope(callback_timeout):
                msg = await future

            if msg:
                if msg.headers:  # pragma: no branch
                    if (
                        msg.headers.get(nats.js.api.Header.STATUS)
                        == nats.aio.client.NO_RESPONDERS_STATUS
                    ):
                        raise nats.errors.NoRespondersError
                return await self._decode_message(await self._parse_message(msg))

    async def close(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_val: Optional[BaseException] = None,
        exec_tb: Optional[TracebackType] = None,
    ) -> None:
        await super().close(exc_type, exc_val, exec_tb)
        for h in self.handlers:
            if h.subscription is not None:
                await h.subscription.unsubscribe()
                h.subscription = None

        if self._connection is not None:
            await self._connection.drain()
            self._connection = None

    def _get_log_context(
        self,
        message: Optional[NatsMessage],
        subject: str,
        queue: str = "",
    ) -> Dict[str, Any]:
        context = {
            "subject": subject,
            "queue": queue,
            **super()._get_log_context(message),
        }
        return context

    @property
    def fmt(self) -> str:
        return self._fmt or (
            "%(asctime)s %(levelname)s - "
            f"%(subject)-{self._max_subject_len}s | "
            + (f"%(queue)-{self._max_queue_len}s | " if self._max_queue_len else "")
            + "%(message_id)-10s "
            "- %(message)s"
        )

    async def _parse_message(self, message: Msg) -> NatsMessage:
        headers = message.header or {}
        return PropanMessage(
            body=message.data,
            content_type=headers.get("content-type", ""),
            headers=headers,
            reply_to=message.reply,
            raw_message=message,
        )

    def _process_message(
        self,
        func: Callable[[NatsMessage], Awaitable[T]],
        watcher: Optional[BaseWatcher] = None,
    ) -> Callable[[NatsMessage], Awaitable[T]]:
        @wraps(func)
        async def wrapper(message: NatsMessage) -> T:
            r = await func(message)

            if message.reply_to:
                await self.publish(r, message.reply_to)

            return r

        return wrapper

    def log_connection_broken(
        self, error_cb: Optional[ErrorCallback] = None
    ) -> ErrorCallback:
        c = self._get_log_context(None, "")

        async def wrapper(err: Exception) -> None:
            if error_cb is not None:
                await error_cb(err)

            if self.__is_connected is True:
                self._log(err, logging.WARNING, c)
                self.__is_connected = False

        return wrapper

    def log_reconnected(self, cb: Optional[Callback] = None) -> Callback:
        c = self._get_log_context(None, "")

        async def wrapper() -> None:
            if cb is not None:
                await cb()

            if self.__is_connected is False:
                self._log("Connection established", logging.INFO, c)
                self.__is_connected = True

        return wrapper



import logging
from abc import ABC
from typing import Awaitable, Callable, Dict, List, Optional

import anyio
from typing_extensions import ParamSpec, Protocol, TypeVar

from propan.asyncapi.info import AsyncAPIContact, AsyncAPILicense
from propan.cli.supervisors.utils import set_exit
from propan.cli.utils.parser import SettingField
from propan.log import logger
from propan.types import AnyCallable, AsyncFunc
from propan.utils import apply_types, context
from propan.utils.functions import to_async


class Runnable(Protocol):
    async def start(self) -> None:
        ...

    async def close(self) -> None:
        ...


P_HookParams = ParamSpec("P_HookParams")
T_HookReturn = TypeVar("T_HookReturn")


class ABCApp(ABC):
    _on_startup_calling: List[AnyCallable]
    _after_startup_calling: List[AnyCallable]
    _on_shutdown_calling: List[AnyCallable]
    _after_shutdown_calling: List[AnyCallable]

    license: Optional[AsyncAPILicense]
    contact: Optional[AsyncAPIContact]

    def __init__(
        self,
        broker: Optional[Runnable] = None,
        logger: Optional[logging.Logger] = logger,
        # AsyncAPI args,
        title: str = "Propan",
        version: str = "0.1.0",
        description: str = "",
        license: Optional[AsyncAPILicense] = None,
        contact: Optional[AsyncAPIContact] = None,
    ):
        self.broker = broker
        self.logger = logger
        self.context = context
        context.set_global("app", self)

        self._on_startup_calling = []
        self._after_startup_calling = []
        self._on_shutdown_calling = []
        self._after_shutdown_calling = []
        self._command_line_options: Dict[str, SettingField] = {}

        self.title = title
        self.version = version
        self.description = description
        self.license = license
        self.contact = contact

    def set_broker(self, broker: Runnable) -> None:
        """Set already existed App object broker
        Usefull then you create/init broker in `on_startup` hook"""
        self.broker = broker

    def on_startup(
        self, func: Callable[P_HookParams, T_HookReturn]
    ) -> Callable[P_HookParams, T_HookReturn]:
        """Add hook running BEFORE broker connected
        This hook also takes an extra CLI options as a kwargs"""
        self._on_startup_calling.append(apply_types(func))
        return func

    def on_shutdown(
        self, func: Callable[P_HookParams, T_HookReturn]
    ) -> Callable[P_HookParams, T_HookReturn]:
        """Add hook running BEFORE broker disconnected"""
        self._on_shutdown_calling.append(apply_types(func))
        return func

    def after_startup(
        self, func: Callable[P_HookParams, T_HookReturn]
    ) -> Callable[P_HookParams, T_HookReturn]:
        """Add hook running AFTER broker connected"""
        self._after_startup_calling.append(apply_types(func))
        return func

    def after_shutdown(
        self, func: Callable[P_HookParams, T_HookReturn]
    ) -> Callable[P_HookParams, T_HookReturn]:
        """Add hook running AFTER broker disconnected"""
        self._after_shutdown_calling.append(apply_types(func))
        return func

    def _log(self, level: int, message: str) -> None:
        if self.logger is not None:
            self.logger.log(level, message)


class PropanApp(ABCApp):
    _on_startup_calling: List[AsyncFunc]
    _after_startup_calling: List[AsyncFunc]
    _on_shutdown_calling: List[AsyncFunc]
    _after_shutdown_calling: List[AsyncFunc]

    _stop_event: Optional[anyio.Event]

    def __init__(
        self,
        broker: Optional[Runnable] = None,
        logger: Optional[logging.Logger] = logger,
        # AsyncAPI args,
        title: str = "Propan",
        version: str = "0.1.0",
        description: str = "",
        license: Optional[AsyncAPILicense] = None,
        contact: Optional[AsyncAPIContact] = None,
    ):
        """Asyncronous Propan Application class

        stores and run broker, control hooks

        Args:
            broker: async broker to run (may be `None`, then specify by `set_broker`)
            logger: logger object to log startup/shutdown messages (`None` to disable)

        AsyncAPI Args:
            title: application title
            version: application version
            description: application description
            license: application license
            contact: application contact
        """
        super().__init__(
            broker=broker,
            logger=logger,
            title=title,
            version=version,
            description=description,
            license=license,
            contact=contact,
        )

        self._stop_event = None

        set_exit(lambda *_: self.__exit())

    def on_startup(
        self, func: Callable[P_HookParams, T_HookReturn]
    ) -> Callable[P_HookParams, Awaitable[T_HookReturn]]:
        """Add hook running BEFORE broker connected

        This hook also takes an extra CLI options as a kwargs

        Args:
            func: async or sync func to call as a hook

        Returns:
            Async version of the func argument
        """
        return super().on_startup(to_async(func))

    def on_shutdown(
        self, func: Callable[P_HookParams, Awaitable[T_HookReturn]]
    ) -> AsyncFunc:
        """Add hook running BEFORE broker disconnected

        Args:
            func: async or sync func to call as a hook

        Returns:
            Async version of the func argument
        """
        return super().on_shutdown(to_async(func))

    def after_startup(
        self, func: Callable[P_HookParams, Awaitable[T_HookReturn]]
    ) -> AsyncFunc:
        """Add hook running AFTER broker connected

        Args:
            func: async or sync func to call as a hook

        Returns:
            Async version of the func argument
        """
        return super().after_startup(to_async(func))

    def after_shutdown(
        self, func: Callable[P_HookParams, Awaitable[T_HookReturn]]
    ) -> AsyncFunc:
        """Add hook running AFTER broker disconnected

        Args:
            func: async or sync func to call as a hook

        Returns:
            Async version of the func argument
        """
        return super().after_shutdown(to_async(func))

    async def run(self, log_level: int = logging.INFO) -> None:
        """Run Propan Application

        Args:
            log_level: force application log level

        Returns:
            Block an event loop until stopped
        """
        self._init_async_cycle()
        async with anyio.create_task_group() as tg:
            tg.start_soon(self._start, log_level)
            await self._stop(log_level)
            tg.cancel_scope.cancel()

    def _init_async_cycle(self) -> None:
        if self._stop_event is None:
            self._stop_event = anyio.Event()

    async def _start(self, log_level: int = logging.INFO) -> None:
        self._log(log_level, "Propan app starting...")
        await self._startup()
        self._log(log_level, "Propan app started successfully! To exit press CTRL+C")

    async def _stop(self, log_level: int = logging.INFO) -> None:
        assert self._stop_event, "You should call `_init_async_cycle` first"
        await self._stop_event.wait()
        self._log(log_level, "Propan app shutting down...")
        await self._shutdown()
        self._log(log_level, "Propan app shut down gracefully.")

    async def _startup(self) -> None:
        for func in self._on_startup_calling:
            await func(**self._command_line_options)

        if self.broker is not None:
            await self.broker.start()

        for func in self._after_startup_calling:
            await func()

    async def _shutdown(self) -> None:
        for func in self._on_shutdown_calling:
            await func()

        if (
            self.broker is not None
            and getattr(self.broker, "_connection", False) is not False
        ):
            await self.broker.close()

        for func in self._after_shutdown_calling:
            await func()

    def __exit(self) -> None:
        if self._stop_event is not None:  # pragma: no branch
            self._stop_event.set()



from typing import Optional

from pydantic import BaseModel, Field


class AsyncAPIExternalDocs(BaseModel):
    url: str = ""
    description: str = ""


class AsyncAPITag(BaseModel):
    name: str
    description: str = ""
    external_docs: Optional[AsyncAPIExternalDocs] = Field(
        default=None,
        alias="externalDocs",
    )



"""
Alias is able to provide access to specific attributes of dependency
"""
from pydantic import BaseSettings, Field

from propan import Context, PropanApp, RabbitBroker
from propan.annotations import ContextRepo

broker = RabbitBroker("amqp://guest:guest@localhost:5672/")

app = PropanApp(broker)

KEY = "my-secret-key"


class Settings(BaseSettings):
    key: str = Field("")


@app.on_startup
async def setup(context: ContextRepo):
    settings = Settings(key=KEY)
    context.set_global("settings", settings)


@app.on_startup
def setup_later(key: str = Context("settings.key")):
    assert key is KEY



from propan import __about__ as about

try:
    from propan.test.rabbit import TestRabbitBroker
except Exception:
    TestRabbitBroker = about.INSTALL_RABBIT

try:
    from propan.test.redis import TestRedisBroker
except Exception:
    TestRedisBroker = about.INSTALL_REDIS

try:
    from propan.test.kafka import TestKafkaBroker
except Exception:
    TestKafkaBroker = about.INSTALL_KAFKA

try:
    from propan.test.nats import TestNatsBroker
except Exception:
    TestNatsBroker = about.INSTALL_NATS

try:
    from propan.test.sqs import TestSQSBroker
except Exception:
    TestSQSBroker = about.INSTALL_SQS

assert any(
    (TestRabbitBroker, TestRedisBroker, TestKafkaBroker, TestNatsBroker, TestSQSBroker)
), about.INSTALL_MESSAGE

__all__ = (
    "TestRabbitBroker",
    "TestRedisBroker",
    "TestKafkaBroker",
    "TestNatsBroker",
    "TestSQSBroker",
)



import logging
from enum import Enum
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    Union,
)

from fastapi import params
from fastapi.datastructures import Default
from fastapi.routing import APIRoute
from fastapi.utils import generate_unique_id
from redis.asyncio.connection import BaseParser, DefaultParser, Encoder
from starlette import routing
from starlette.responses import JSONResponse, Response
from starlette.types import ASGIApp

from propan import RedisBroker
from propan.brokers._model.broker_usecase import (
    AsyncDecoder,
    AsyncParser,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.fastapi.core.router import PropanRouter
from propan.log import access_logger
from propan.types import AnyDict

class RedisRouter(PropanRouter[RedisBroker]):
    def __init__(
        self,
        url: str = "redis://localhost:6379",
        polling_interval: float = 1.0,
        host: str = "localhost",
        port: Union[str, int] = 6379,
        db: Union[str, int] = 0,
        password: Optional[str] = None,
        socket_timeout: Optional[float] = None,
        socket_connect_timeout: Optional[float] = None,
        socket_keepalive: bool = False,
        socket_keepalive_options: Optional[Mapping[int, Union[int, bytes]]] = None,
        socket_type: int = 0,
        retry_on_timeout: bool = False,
        encoding: str = "utf-8",
        encoding_errors: str = "strict",
        decode_responses: bool = False,
        parser_class: Type[BaseParser] = DefaultParser,
        socket_read_size: int = 65536,
        health_check_interval: float = 0,
        client_name: Optional[str] = None,
        username: Optional[str] = None,
        encoder_class: Type[Encoder] = Encoder,
        # FastAPI kwargs
        prefix: str = "",
        tags: Optional[List[Union[str, Enum]]] = None,
        dependencies: Optional[Sequence[params.Depends]] = None,
        default_response_class: Type[Response] = Default(JSONResponse),
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        callbacks: Optional[List[routing.BaseRoute]] = None,
        routes: Optional[List[routing.BaseRoute]] = None,
        redirect_slashes: bool = True,
        default: Optional[ASGIApp] = None,
        dependency_overrides_provider: Optional[Any] = None,
        route_class: Type[APIRoute] = APIRoute,
        on_startup: Optional[Sequence[Callable[[], Any]]] = None,
        on_shutdown: Optional[Sequence[Callable[[], Any]]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        generate_unique_id_function: Callable[[APIRoute], str] = Default(
            generate_unique_id
        ),
        # Broker kwargs
        schema_url: str = "/asyncapi",
        logger: Optional[logging.Logger] = access_logger,
        log_level: int = logging.INFO,
        log_fmt: Optional[str] = None,
        apply_types: bool = True,
        decode_message: AsyncDecoder[AnyDict] = None,
        parse_message: AsyncParser[AnyDict] = None,
        protocol: str = "redis",
    ) -> None:
        pass
    def add_api_mq_route(  # type: ignore[override]
        self,
        channel: str,
        *,
        endpoint: HandlerCallable[T_HandlerReturn],
        pattern: bool = False,
        decode_message: AsyncDecoder[AnyDict] = None,
        parse_message: AsyncParser[AnyDict] = None,
        description: str = "",
        dependencies: Optional[Sequence[params.Depends]] = None,
    ) -> Callable[[AnyDict, bool], Awaitable[T_HandlerReturn]]:
        pass
    def event(  # type: ignore[override]
        self,
        channel: str,
        *,
        pattern: bool = False,
        decode_message: AsyncDecoder[AnyDict] = None,
        parse_message: AsyncParser[AnyDict] = None,
        description: str = "",
        dependencies: Optional[Sequence[params.Depends]] = None,
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[AnyDict, bool], Awaitable[T_HandlerReturn]],
    ]:
        pass



from propan.asyncapi.bindings import AsyncAPIChannelBinding
from propan.asyncapi.channels import AsyncAPIChannel
from propan.asyncapi.info import AsyncAPIContact, AsyncAPIInfo, AsyncAPILicense
from propan.asyncapi.main import ASYNC_API_VERSION, AsyncAPIComponents, AsyncAPISchema
from propan.asyncapi.message import AsyncAPIMessage
from propan.asyncapi.security import AsyncAPISecuritySchemeComponent
from propan.asyncapi.servers import AsyncAPIServer
from propan.asyncapi.utils import AsyncAPIExternalDocs, AsyncAPITag

__all__ = (
    # main
    "ASYNC_API_VERSION",
    "AsyncAPISchema",
    "AsyncAPIComponents",
    # info
    "AsyncAPIInfo",
    "AsyncAPIContact",
    "AsyncAPILicense",
    # servers
    "AsyncAPIServer",
    # channels
    "AsyncAPIChannel",
    # utils
    "AsyncAPITag",
    "AsyncAPIExternalDocs",
    # bindings
    "AsyncAPIChannelBinding",
    # messages
    "AsyncAPIMessage",
    # security
    "AsyncAPISecuritySchemeComponent",
)



import json
from typing import Any, Callable, Dict, Optional, Union

from propan.asyncapi import AsyncAPISchema
from propan.cli.docs.gen import schema_to_json


def serve_docs(
    schema: str = "",
    host: str = "0.0.0.0",
    port: int = 8000,
    raw_schema: Optional[AsyncAPISchema] = None,
) -> None:
    if not any((schema, raw_schema)):
        raise ValueError("You should set `shema` or `raw_schema`")

    import uvicorn
    from fastapi import FastAPI

    app = FastAPI()

    app.get("/")(asyncapi_html_endpoint(schema, raw_schema))

    if raw_schema is not None:
        app.get("/asyncapi.json")(download_json_endpoint(raw_schema))

    app.get("/asyncapi.yaml")(download_yaml_endpoint(schema))
    uvicorn.run(app, host=host, port=port)


def download_yaml_endpoint(schema: str) -> Callable[[], Any]:
    from fastapi.responses import Response

    def download_yaml() -> Response:
        return Response(
            content=schema,
            headers={
                "Content-Type": "application/octet-stream",
            },
        )

    return download_yaml


def download_json_endpoint(raw_schema: AsyncAPISchema) -> Callable[[], Any]:
    from fastapi.responses import Response

    def download_json() -> Response:
        return Response(
            content=json.dumps(
                schema_to_json(raw_schema),
                indent=4,
            ),
            headers={
                "Content-Type": "application/octet-stream",
            },
        )

    return download_json


def asyncapi_html_endpoint(
    schema: str, raw_schema: Optional[AsyncAPISchema] = None
) -> Callable[[], Any]:
    from fastapi.responses import HTMLResponse

    def asyncapi(
        sidebar: bool = True,
        info: bool = True,
        servers: bool = True,
        operations: bool = True,
        messages: bool = True,
        schemas: bool = True,
        errors: bool = True,
        expandMessageExamples: bool = True,
    ) -> HTMLResponse:
        return HTMLResponse(
            content=get_asyncapi_html(
                schema,
                sidebar=sidebar,
                info=info,
                servers=servers,
                operations=operations,
                messages=messages,
                schemas=schemas,
                errors=errors,
                expand_message_examples=expandMessageExamples,
                title=raw_schema.info.title if raw_schema else "Propan",
            )
        )

    return asyncapi


def get_asyncapi_html(
    schema: Union[str, Dict[str, Any]],
    sidebar: bool = True,
    info: bool = True,
    servers: bool = True,
    operations: bool = True,
    messages: bool = True,
    schemas: bool = True,
    errors: bool = True,
    expand_message_examples: bool = True,
    title: str = "Propan",
) -> str:
    config = {
        "schema": schema,
        "config": {
            "show": {
                "sidebar": sidebar,
                "info": info,
                "servers": servers,
                "operations": operations,
                "messages": messages,
                "schemas": schemas,
                "errors": errors,
            },
            "expand": {
                "messageExamples": expand_message_examples,
            },
            "sidebar": {
                "showServers": "byDefault",
                "showOperations": "byDefault",
            },
        },
    }

    return (
        """
    <!DOCTYPE html>
    <html>
        <head>
    """
        f"""
        <title>{title} AsyncAPI</title>
    """
        """
        <link rel="icon" href="https://www.asyncapi.com/favicon.ico">
        <link rel="icon" type="image/png" sizes="16x16" href="https://www.asyncapi.com/favicon-16x16.png">
        <link rel="icon" type="image/png" sizes="32x32" href="https://www.asyncapi.com/favicon-32x32.png">
        <link rel="icon" type="image/png" sizes="194x194" href="https://www.asyncapi.com/favicon-194x194.png">
        <link rel="stylesheet" href="https://unpkg.com/@asyncapi/react-component@1.0.0-next.46/styles/default.min.css">
        </head>

        <style>
        html {
            font-family: ui-sans-serif,system-ui,-apple-system,BlinkMacSystemFont,Segoe UI,Roboto,Helvetica Neue,Arial,Noto Sans,sans-serif,Apple Color Emoji,Segoe UI Emoji,Segoe UI Symbol,Noto Color Emoji;
            line-height: 1.5;
        }
        </style>

        <body>
        <div id="asyncapi"></div>

        <script src="https://unpkg.com/@asyncapi/react-component@1.0.0-next.47/browser/standalone/index.js"></script>
        <script>
    """
        f"""
            AsyncApiStandalone.render({json.dumps(config)}, document.getElementById('asyncapi'));
    """
        """
        </script>
        </body>
    </html>
    """
    )



from typing import Optional

from pydantic import BaseModel, Field
from typing_extensions import Literal

from propan.types import AnyDict


class AsyncAPIAmqpQueue(BaseModel):
    name: str
    durable: bool
    exclusive: bool
    auto_delete: bool = Field(alias="autoDelete")
    vhost: str = "/"


class AsyncAPIAmqpExchange(BaseModel):
    name: Optional[str] = None
    type: Literal["default", "direct", "topic", "fanout", "headers"]
    durable: Optional[bool] = None
    auto_delete: Optional[bool] = Field(
        default=None,
        alias="autoDelete",
    )
    vhost: str = "/"


class AsyncAPIAmqpChannelBinding(BaseModel):
    is_: Literal["queue", "routingKey"] = Field(..., alias="is")
    version: str = Field(
        default="0.2.0",
        alias="bindingVersion",
    )
    queue: Optional[AsyncAPIAmqpQueue] = None
    exchange: Optional[AsyncAPIAmqpExchange] = None


class AsyncAPIAmqpOperationBinding(BaseModel):
    cc: Optional[str] = None
    ack: bool = True
    reply_to: Optional[AnyDict] = Field(default=None, alias="replyTo")
    version: str = Field(
        default="0.2.0",
        alias="bindingVersion",
    )



import asyncio

from propan.brokers.rabbit import RabbitBroker

broker = RabbitBroker("amqp://guest:guest@127.0.0.1/")


@broker.handle("test")
async def ping(m: str):
    return "hello!"


async def main():
    async with broker:
        r = await broker.publish(
            "hello",
            routing_key="test",
            callback=True,
        )
        assert r == "hello"


asyncio.run(main())



from pathlib import Path

from fast_depends._compat import PYDANTIC_V2

from propan.__about__ import __version__
from propan.cli.startproject.utils import touch_dir, write_file


def create_project_dir(dirname: Path, version: str) -> Path:
    project_dir = touch_dir(dirname)
    create_readme(project_dir)
    create_dockerfile(project_dir)
    create_requirements(project_dir, version)
    create_gitignore(project_dir)
    return project_dir


def create_readme(project_dir: Path) -> None:
    write_file(project_dir / "README.md")


def create_dockerfile(project_dir: Path) -> None:
    write_file(
        project_dir / "Dockerfile",
        "FROM python:3.11.3-slim",
        "",
        "ENV PYTHONUNBUFFERED=1 \\",
        "    PYTHONDONTWRITEBYTECODE=1 \\",
        "    PIP_DISABLE_PIP_VERSION_CHECK=on",
        "",
        "RUN useradd -ms /bin/bash user",
        "",
        "USER user",
        "WORKDIR /home/user",
        "",
        "COPY requirements.txt requirements.txt",
        "",
        "RUN pip install --no-warn-script-location --no-cache-dir -r requirements.txt",
        "",
        "COPY ./app ./app",
        "",
        'ENTRYPOINT [ "python", "-m", "propan", "run", "app.serve:app" ]',
    )


def create_gitignore(project_dir: Path) -> None:
    write_file(
        project_dir / ".gitignore",
        "*.env*",
        "__pycache__",
        "env/",
        "venv/",
    )


def create_requirements(project_dir: Path, version: str) -> None:
    write_file(
        project_dir / "requirements.txt",
        f"{version}=={__version__}",
        "python-dotenv",
        "pydantic-settings>=2.0a1" if PYDANTIC_V2 else "",
    )


def create_env(config_dir: Path, url: str) -> None:
    write_file(
        config_dir / ".env",
        "APP_DEBUG=True",
        f"APP_BROKER__URL={url}",
    )


def create_config_dir(config: Path, *broker_settings: str) -> Path:
    if not broker_settings:
        broker_settings = (
            "class BrokerSettings(BaseModel):",
            "    url: str = Field(...)",
        )

    config_dir = touch_dir(config)

    write_file(
        config_dir / "settings.py",
        "from pathlib import Path",
        "from typing import Optional",
        "",
        "from pydantic import BaseModel, Field",
        "from pydantic_settings import BaseSettings"
        if PYDANTIC_V2
        else "from pydantic import BaseSettings",  # TODO: remove it from with stable PydanticV2
        "",
        "",
        "CONFIG_DIR = Path(__file__).resolve().parent",
        "BASE_DIR = CONFIG_DIR.parent",
        "",
        "",
        *broker_settings,
        "",
        "",
        "class Settings(BaseSettings):",
        "    debug: bool = Field(True)",
        "    broker: BrokerSettings = Field(default_factory=BrokerSettings)",
        "    base_dir: Path = BASE_DIR",
        "",
        "    class Config:",
        "        env_prefix = 'APP_'",
        "        env_file_encoding = 'utf-8'",
        "        env_nested_delimiter = '__'",
        "",
        "",
        "def init_settings(env_file: Optional[str] = None) -> Settings:",
        '    env_file = CONFIG_DIR / (env_file or ".env")',
        "    if env_file.exists():",
        "        settings = Settings(_env_file=env_file)",
        "    else:",
        "        settings = Settings()",
        "    return settings",
    )

    write_file(
        config_dir / "__init__.py",
        "from config.settings import Settings, init_settings",
    )

    return config_dir


def create_apps_dir(apps: Path) -> Path:
    apps_dir = touch_dir(apps)

    write_file(
        apps_dir / "__init__.py",
        "from apps.handlers import router",
    )

    return apps_dir



# Imports to use at __all__
from propan import __about__ as about
from propan.brokers import PropanMessage
from propan.brokers.middlewares import BaseMiddleware
from propan.cli.app import *  # noqa: F403
from propan.log import *  # noqa: F403
from propan.utils import *  # noqa: F403

try:
    from propan.brokers.rabbit import RabbitBroker, RabbitRouter
except ImportError:
    RabbitBroker = RabbitRouter = about.INSTALL_RABBIT  # type: ignore

try:
    from propan.brokers.nats import NatsBroker, NatsJSBroker, NatsRouter
except ImportError:
    NatsJSBroker = NatsBroker = NatsRouter = about.INSTALL_NATS  # type: ignore

try:
    from propan.brokers.redis import RedisBroker, RedisRouter
except ImportError:
    RedisBroker = RedisRouter = about.INSTALL_REDIS  # type: ignore

try:
    from propan.brokers.kafka import KafkaBroker, KafkaRouter
except ImportError:
    KafkaBroker = KafkaRouter = about.INSTALL_KAFKA  # type: ignore

try:
    from propan.brokers.sqs import SQSBroker, SQSRouter
except ImportError:
    SQSBroker = SQSRouter = about.INSTALL_SQS  # type: ignore

assert any(
    (RabbitBroker, NatsBroker, RedisBroker, SQSBroker, KafkaBroker)
), about.INSTALL_MESSAGE

__all__ = (  # noqa: F405
    # app
    "PropanApp",
    # log
    "logger",
    "access_logger",
    # utils
    ## types
    "apply_types",
    ## context
    "context",
    "Context",
    "ContextRepo",
    "Depends",
    # brokers
    "PropanMessage",
    "BaseMiddleware",
    ## nats
    "NatsBroker",
    "NatsJSBroker",
    "NatsRouter",
    ## rabbit
    "RabbitBroker",
    "RabbitRouter",
    ## redis
    "RedisBroker",
    "RedisRouter",
    ## kafka
    "KafkaRouter",
    "KafkaBroker",
    ## sqs
    "SQSBroker",
    "SQSRouter",
)



from typing import Awaitable, Callable, Union

from propan._compat import model_copy
from propan.brokers._model.broker_usecase import HandlerCallable, T_HandlerReturn
from propan.brokers._model.routing import BrokerRouter
from propan.brokers.sqs.schema import SQSQueue
from propan.types import AnyDict


class SQSRouter(BrokerRouter[AnyDict]):
    def handle(  # type: ignore[override]
        self,
        queue: Union[str, SQSQueue],
        **kwargs: AnyDict,
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[AnyDict, bool], Awaitable[T_HandlerReturn]],
    ]:
        if isinstance(queue, str):
            queue = SQSQueue(queue)

        return super().handle(model_copy(queue, update={
            "name": self.prefix + queue.name
        }), **kwargs)



import asyncio

from propan.brokers.rabbit import RabbitBroker

broker = RabbitBroker("amqp://guest:guest@127.0.0.1/")


@broker.handle("test")
async def ping(m: str):
    return "hello!"


@broker.handle("reply")
async def get_message(m: str):
    assert m == "hello!"


async def main():
    await broker.start()

    await broker.publish("hi!", queue="test", reply_to="reply")

    try:
        await asyncio.Future()
    finally:
        await broker.close()


asyncio.run(main())



from propan import PropanApp, RabbitBroker
from propan.annotations import Logger
from propan.brokers.rabbit import ExchangeType, RabbitExchange, RabbitQueue

broker = RabbitBroker()
app = PropanApp(broker)

header_exchange = RabbitExchange(
    "exchange", auto_delete=True, type=ExchangeType.HEADERS
)

queue_1 = RabbitQueue("test-queue-1", auto_delete=True, bind_arguments={"key": 1})
queue_2 = RabbitQueue(
    "test-queue-2",
    auto_delete=True,
    bind_arguments={"key": 2, "key2": 2, "x-match": "any"},
)
queue_3 = RabbitQueue(
    "test-queue-3",
    auto_delete=True,
    bind_arguments={"key": 2, "key2": 2, "x-match": "all"},
)


@broker.handle(queue_1, header_exchange)
async def base_handler1(logger: Logger):
    logger.info("base_handler1")


@broker.handle(queue_1, header_exchange)
async def base_handler2(logger: Logger):
    logger.info("base_handler2")


@broker.handle(queue_2, header_exchange)
async def base_handler3(logger: Logger):
    logger.info("base_handler3")


@broker.handle(queue_3, header_exchange)
async def base_handler4(logger: Logger):
    logger.info("base_handler4")


@app.after_startup
async def send_messages():
    await broker.publish(exchange=header_exchange, headers={"key": 1})
    await broker.publish(exchange=header_exchange, headers={"key": 1})
    await broker.publish(exchange=header_exchange, headers={"key": 1})
    await broker.publish(exchange=header_exchange, headers={"key": 2})
    await broker.publish(exchange=header_exchange, headers={"key2": 2})
    await broker.publish(exchange=header_exchange, headers={"key": 2, "key2": 2})



from propan.brokers.sqs.routing import SQSRouter
from propan.brokers.sqs.schema import (
    FifoQueue,
    RedriveAllowPolicy,
    RedrivePolicy,
    SQSQueue,
)
from propan.brokers.sqs.sqs_broker import SQSBroker, SQSMessage

__all__ = (
    "SQSBroker",
    "SQSQueue",
    "SQSRouter",
    "RedrivePolicy",
    "RedriveAllowPolicy",
    "FifoQueue",
    "SQSMessage",
)



from fastapi import Depends, FastAPI
from pydantic import BaseModel

from propan.fastapi import RabbitRouter

router = RabbitRouter("amqp://guest:guest@localhost:5672")

app = FastAPI(lifespan=router.lifespan_context)


class Incoming(BaseModel):
    m: dict


def call():
    return True


@router.event("test")
async def hello(m: Incoming, d=Depends(call)) -> dict:
    return {"response": "Hello, world!"}


@router.get("/")
async def hello_http():
    return "Hello, http!"


app.include_router(router)



import logging

from typing_extensions import Annotated, TypeVar

from propan import __about__ as about
from propan.cli.app import PropanApp
from propan.utils.context import Context as ContextField
from propan.utils.context import ContextRepo as CR
from propan.utils.no_cast import NoCast as NC

Logger = Annotated[logging.Logger, ContextField("logger")]
App = Annotated[PropanApp, ContextField("app")]
ContextRepo = Annotated[CR, ContextField("context")]

NoCastType = TypeVar("NoCastType")
NoCast = Annotated[NoCastType, NC()]

try:
    import aio_pika

    from propan.brokers.rabbit import RabbitBroker as RB

    RabbitBroker = Annotated[RB, ContextField("broker")]
    RabbitMessage = Annotated[aio_pika.message.IncomingMessage, ContextField("message")]
    Channel = Annotated[
        aio_pika.robust_channel.RobustChannel, ContextField("broker.channel")
    ]
except ImportError:
    RabbitBroker = RabbitMessage = Channel = about.INSTALL_RABBIT


try:
    from nats.aio.client import Client
    from nats.aio.msg import Msg
    from nats.js.client import JetStreamContext

    from propan.brokers.nats import NatsBroker as NB
    from propan.brokers.nats import NatsJSBroker as NJB

    NatsBroker = Annotated[NB, ContextField("broker")]
    NatsJSBroker = Annotated[NJB, ContextField("broker")]
    NatsMessage = Annotated[Msg, ContextField("message")]
    NatsConnection = Annotated[Client, ContextField("broker._connection")]
    NatsJS = Annotated[JetStreamContext, ContextField("broker._connection")]
except ImportError:
    NatsBroker = (
        NatsMessage
    ) = NatsJSBroker = NatsJS = NatsConnection = about.INSTALL_NATS


try:
    from redis.asyncio.client import Redis as R

    from propan.brokers.redis import RedisBroker as RedB

    RedisBroker = Annotated[RedB, ContextField("broker")]
    Redis = Annotated[R, ContextField("broker._connection")]
except ImportError:
    RedisBroker = Redis = about.INSTALL_REDIS


try:
    from aiokafka import AIOKafkaProducer
    from aiokafka.structs import ConsumerRecord

    from propan.brokers.kafka import KafkaBroker as KB

    KafkaBroker = Annotated[KB, ContextField("broker")]
    KafkaMessage = Annotated[ConsumerRecord, ContextField("message")]
    Producer = Annotated[AIOKafkaProducer, ContextField("producer")]
except ImportError:
    KafkaBroker = KafkaMessage = Producer = about.INSTALL_KAFKA


try:
    from aiobotocore.client import AioBaseClient

    from propan.brokers.sqs import SQSBroker as SB

    SQSBroker = Annotated[SB, ContextField("broker")]
    client = Annotated[AioBaseClient, ContextField("client")]
    queue_url = Annotated[str, ContextField("queue_url")]
except ImportError:
    SQSBroker = client = queue_url = about.INSTALL_SQS


assert any(
    (
        all((RabbitBroker, RabbitMessage, Channel)),
        all((NatsBroker, NatsJSBroker, NatsJS, NatsMessage)),
        all((RedisBroker, Redis)),
        all((KafkaBroker, KafkaMessage, Producer)),
        all((SQSBroker, client, queue_url)),
    )
), about.INSTALL_MESSAGE



import asyncio
from dataclasses import dataclass
from dataclasses import field as DField
from typing import Any, Dict, Optional, Sequence

from fast_depends.core import CallModel
from pydantic import BaseModel, Field, PositiveInt
from typing_extensions import Literal

from propan._compat import model_to_dict
from propan.asyncapi.bindings import (
    AsyncAPIChannelBinding,
    AsyncAPIOperationBinding,
    sqs,
)
from propan.asyncapi.channels import AsyncAPIChannel
from propan.asyncapi.message import AsyncAPICorrelationId, AsyncAPIMessage
from propan.asyncapi.subscription import AsyncAPISubscription
from propan.brokers._model import BrokerAsyncUsecase
from propan.brokers._model.schemas import BaseHandler, Queue
from propan.types import AnyDict, DecoratedCallable, SendableMessage


class RedrivePolicy(BaseModel):
    """SQS Queue RedrivePolicy attribute details"""

    dead_letter_target: str = Field(
        default="",
        alias="deadLetterTargetArn",
    )
    max_receive_count: PositiveInt = Field(
        default=10,
        alias="deadLetterTargetArn",
    )


class RedriveAllowPolicy(BaseModel):
    """SQS Queue RedriveAllowPolicy attribute details"""

    redrive_permission: Literal["allowAll", "denyAll", "byQueue"] = Field(
        default="allowAll",
        alias="redrivePermission",
    )
    source_queue_arns: Sequence[str] = Field(
        default_factory=tuple,
        alias="sourceQueueArns",
        max_length=10,
    )


class SQSQueue(Queue):
    """SQS Basic Queue initialization attributes"""

    fifo: bool = Field(
        default=False,
        alias="FifoQueue",
    )

    delay_seconds: int = Field(
        default=0,
        # alias="DelaySeconds",
        ge=0,
        le=900,
    )
    max_message_size: int = Field(
        default=262_144,
        alias="MaximumMessageSize",
        ge=1024,
        le=262_144,
    )
    retention_period_sec: int = Field(
        345_600,
        alias="MessageRetentionPeriod",
        ge=60,
        le=1_209_600,
    )
    receive_wait_time_sec: int = Field(
        default=0,
        alias="ReceiveMessageWaitTimeSeconds",
        ge=0,
        le=20,
    )
    visibility_timeout_sec: int = Field(
        default=30,
        alias="VisibilityTimeout",
        ge=0,
        le=43_200,
    )
    redrive_policy: RedrivePolicy = Field(
        default_factory=RedrivePolicy,
        alias="RedrivePolicy",
    )
    redrive_allow_policy: RedriveAllowPolicy = Field(
        default_factory=RedrivePolicy,
        alias="RedriveAllowPolicy",
    )

    kms_master_key_id: str = Field(default="", alias="KmsMasterKeyId")
    kms_data_key_reuse_period_sec: int = Field(
        default=300,
        alias="KmsDataKeyReusePeriodSeconds",
        ge=60,
        le=86_400,
    )
    sse_enabled: bool = Field(
        default=False,
        alias="SqsManagedSseEnabled",
    )
    tags: Dict[str, str] = Field(
        default_factory=dict,
    )

    def __init__(
        self,
        name: str,
        fifo: bool = False,
        delay_seconds: int = 0,
        max_message_size: int = 262_144,
        visibility_timeout_sec: int = 0,
        receive_wait_time_sec: int = 0,
        retention_period_sec: int = 345_600,
        redrive_policy: Optional[RedrivePolicy] = None,
        redrive_allow_policy: Optional[RedriveAllowPolicy] = None,
        kms_master_key_id: str = "",
        kms_data_key_reuse_period_sec: int = 300,
        sse_enabled: bool = False,
        tags: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ):
        super().__init__(
            name=name,
            fifo=fifo,
            visibility_timeout_sec=visibility_timeout_sec,
            receive_wait_time_sec=receive_wait_time_sec,
            retention_period_sec=retention_period_sec,
            max_message_size=max_message_size,
            delay_seconds=delay_seconds,
            redrive_policy=redrive_policy or RedrivePolicy(),
            redrive_allow_policy=redrive_allow_policy or RedriveAllowPolicy(),
            kms_master_key_id=kms_master_key_id,
            kms_data_key_reuse_period_sec=kms_data_key_reuse_period_sec,
            sse_enabled=sse_enabled,
            tags=tags or {},
            **kwargs,
        )


class FifoQueue(SQSQueue):
    """SQS FIFO Queue initialization attributes"""

    fifo: bool = Field(
        default=True,
        alias="FifoQueue",
    )
    content_based_deduplication: bool = Field(
        default=True,
        alias="ContentBasedDeduplication",
    )
    deduplication_scope: Optional[Literal["messageGroup", "queue"]] = Field(
        default=None,
        alias="DeduplicationScope",
    )

    # TODO: pydantic validation and test
    # allow perMessageGroup only for messageGroup deduplication_scope
    throughput_limit: Optional[Literal["perMessageGroup", "perQueue"]] = Field(
        default=None,
        alias="FifoThroughputLimit",
    )

    def __init__(
        self,
        name: str,
        fifo: bool = True,
        delay_seconds: int = 0,
        max_message_size: int = 262_144,
        visibility_timeout_sec: int = 0,
        receive_wait_time_sec: int = 0,
        retention_period_sec: int = 345_600,
        content_based_deduplication: bool = True,
        deduplication_scope: Optional[Literal["messageGroup", "queue"]] = None,
        throughput_limit: Optional[Literal["perMessageGroup", "perQueue"]] = None,
        redrive_policy: Optional[RedrivePolicy] = None,
        redrive_allow_policy: Optional[RedriveAllowPolicy] = None,
        kms_data_key_reuse_period_sec: int = 300,
        sse_enabled: bool = False,
        tags: Optional[Dict[str, str]] = None,
    ):
        super().__init__(
            name=name,
            fifo=fifo,
            visibility_timeout_sec=visibility_timeout_sec,
            receive_wait_time_sec=receive_wait_time_sec,
            content_based_deduplication=content_based_deduplication,
            retention_period_sec=retention_period_sec,
            max_message_size=max_message_size,
            delay_seconds=delay_seconds,
            redrive_policy=redrive_policy or RedrivePolicy(),
            redrive_allow_policy=redrive_allow_policy or RedriveAllowPolicy(),
            kms_data_key_reuse_period_sec=kms_data_key_reuse_period_sec,
            sse_enabled=sse_enabled,
            deduplication_scope=deduplication_scope,
            throughput_limit=throughput_limit,
            tags=tags or {},
        )


@dataclass
class Handler(BaseHandler):
    """Store SQSBroker hanlder information and runtime objects"""

    queue: SQSQueue
    consumer_params: AnyDict

    task: Optional["asyncio.Task[Any]"] = None

    def __init__(
        self,
        callback: DecoratedCallable,
        dependant: CallModel,
        queue: SQSQueue,
        consumer_params: AnyDict,
        _description: str = "",
        task: Optional["asyncio.Task[Any]"] = None,
    ):
        self.callback = callback
        self.dependant = dependant
        self._description = _description
        self.queue = queue
        self.consumer_params = consumer_params
        self.task = task

    def get_schema(self) -> Dict[str, AsyncAPIChannel]:
        message_title, body, reply_to = self.get_message_object()

        return {
            self.title: AsyncAPIChannel(
                subscribe=AsyncAPISubscription(
                    description=self.description,
                    bindings=AsyncAPIOperationBinding(
                        sqs=sqs.AsyncAPISQSOperationBinding(
                            replyTo=reply_to,
                        ),
                    ),
                    message=AsyncAPIMessage(
                        title=message_title,
                        correlationId=AsyncAPICorrelationId(
                            location="$message.header#/correlation_id"
                        ),
                        payload=body,
                    ),
                ),
                bindings=AsyncAPIChannelBinding(
                    sqs=sqs.AsyncAPISQSChannelBinding(
                        queue=model_to_dict(self.queue, include={"name", "fifo"}),
                    )
                ),
            ),
        }


@dataclass
class SQSMessage:
    message: SendableMessage
    delay_seconds: int = 0
    deduplication_id: Optional[str] = None
    group_id: Optional[str] = None
    headers: Dict[str, str] = DField(default_factory=dict)
    message_attributes: AnyDict = DField(default_factory=dict)
    message_system_attributes: AnyDict = DField(default_factory=dict)

    def to_params(self, **extra_headers: Any) -> AnyDict:
        msg, content_type = BrokerAsyncUsecase._encode_message(self.message)

        headers = {**extra_headers, "content-type": content_type, **self.headers}

        params = {
            "MessageBody": msg.decode(),
            "DelaySeconds": self.delay_seconds,
            "MessageSystemAttributes": self.message_system_attributes,
            "MessageAttributes": {
                **self.message_attributes,
                **{
                    i: {
                        "StringValue": j,
                        "DataType": "String",
                    }
                    for i, j in headers.items()
                    if j
                },
            },
        }

        if self.deduplication_id is not None:
            params["MessageDeduplicationId"] = self.deduplication_id

        if self.group_id is not None:
            params["MessageGroupId"] = self.group_id

        return params



import logging
from ssl import SSLContext
from types import TracebackType
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

import aio_pika
import aiormq
from aio_pika.message import IncomingMessage
from fast_depends.dependencies import Depends
from pamqp.common import FieldTable
from typing_extensions import TypeAlias
from yarl import URL

from propan.brokers._model import BrokerAsyncUsecase
from propan.brokers._model.broker_usecase import AsyncDecoder, AsyncParser
from propan.brokers._model.schemas import PropanMessage
from propan.brokers.middlewares import BaseMiddleware
from propan.brokers.push_back_watcher import BaseWatcher
from propan.brokers.rabbit.schemas import Handler, RabbitExchange, RabbitQueue
from propan.log import access_logger
from propan.types import AnyDict, DecodedMessage, SendableMessage

PikaSendableMessage: TypeAlias = Union[aio_pika.message.Message, SendableMessage]
T_HandlerReturn = TypeVar("T_HandlerReturn", bound=PikaSendableMessage)
HandlerCallable: TypeAlias = Callable[
    ..., Union[T_HandlerReturn, Awaitable[T_HandlerReturn]]
]

RabbitMessage: TypeAlias = PropanMessage[IncomingMessage]

class RabbitBroker(BrokerAsyncUsecase[IncomingMessage, aio_pika.RobustConnection]):
    handlers: List[Handler]
    middlewares: Sequence[Type[BaseMiddleware[IncomingMessage]]]
    _channel: Optional[aio_pika.RobustChannel]

    __max_queue_len: int
    __max_exchange_len: int

    def __init__(
        self,
        url: Union[str, URL, None] = None,
        *,
        host: str = "localhost",
        port: int = 5672,
        login: str = "guest",
        password: str = "guest",
        virtualhost: str = "/",
        ssl: bool = False,
        ssl_options: Optional[aio_pika.abc.SSLOptions] = None,
        ssl_context: Optional[SSLContext] = None,
        timeout: aio_pika.abc.TimeoutType = None,
        client_properties: Optional[FieldTable] = None,
        # broker
        logger: Optional[logging.Logger] = access_logger,
        log_level: int = logging.INFO,
        log_fmt: Optional[str] = None,
        apply_types: bool = True,
        consumers: Optional[int] = None,
        dependencies: Sequence[Depends] = (),
        middlewares: Sequence[Type[BaseMiddleware[IncomingMessage]]] = (),
        decode_message: AsyncDecoder[IncomingMessage] = None,
        parse_message: AsyncParser[IncomingMessage] = None,
        # AsyncAPI
        protocol: str = "amqp",
        protocol_version: str = "0.9.1",
    ) -> None:
        """RabbitMQ Propan broker

        URL string might be contain ssl parameters e.g.
        `amqps://user:pass@host:port/vhost?ca_certs=ca.pem&certfile=crt.pem&keyfile=key.pem`

        Args:
            url: RFC3986_ formatted broker address. If `None`
                 will be used keyword arguments.
            host: broker hostname or ip address
            port: broker port
            login: username string.
            password: password string.
            virtualhost: virtualhost parameter.
            client_properties: custom client capability.
            ssl: use SSL for connection. Should be used with addition kwargs.
            ssl_options: A dict of values for the SSL connection.
            timeout: connection timeout in seconds
            ssl_context: ssl.SSLContext instance
            logger: logger to use inside broker
            log_level: broker inner messages log level
            log_fmt: custom log formatting string
            apply_types: wrap brokers handlers to FastDepends decorator
            consumers: max messages to proccess at the same time
            dependencies: dependencies applied to all broker hadlers
            decode_message: custom RabbitMessage decoder
            parse_message: custom IncomingMessage to RabbitMessage parser

        .. _RFC3986: https://goo.gl/MzgYAs
        .. _official Python documentation: https://goo.gl/pty9xA
        """
    async def connect(
        self,
        *,
        url: Union[str, URL, None] = None,
        host: str = "localhost",
        port: int = 5672,
        login: str = "guest",
        password: str = "guest",
        virtualhost: str = "/",
        ssl: bool = False,
        ssl_options: Optional[aio_pika.abc.SSLOptions] = None,
        ssl_context: Optional[SSLContext] = None,
        timeout: aio_pika.abc.TimeoutType = None,
        client_properties: Optional[FieldTable] = None,
    ) -> aio_pika.RobustConnection:
        """Connect to RabbitMQ

        URL string might be contain ssl parameters e.g.
        `amqps://user:pass@host:port/vhost?ca_certs=ca.pem&certfile=crt.pem&keyfile=key.pem`

        Args:
            url: RFC3986_ formatted broker address. If `None`
                 will be used keyword arguments.
            host: broker hostname or ip address
            port: broker port
            login: username string.
            password: password string.
            virtualhost: virtualhost parameter.
            client_properties: custom client capability.
            ssl: use SSL for connection. Should be used with addition kwargs.
            ssl_options: A dict of values for the SSL connection.
            timeout: connection timeout in seconds
            ssl_context: ssl.SSLContext instance

        Returns:
            aio_pika.Connection object

        _RFC3986: https://goo.gl/MzgYAs
        _official Python documentation: https://goo.gl/pty9xA
        """
    async def publish(  # type: ignore[override]
        self,
        message: PikaSendableMessage = "",
        queue: Union[RabbitQueue, str] = "",
        exchange: Union[RabbitExchange, str, None] = None,
        *,
        # publish kwargs
        routing_key: str = "",
        mandatory: bool = True,
        immediate: bool = False,
        timeout: aio_pika.abc.TimeoutType = None,
        # callback kwargs
        callback: bool = False,
        callback_timeout: Optional[float] = 30.0,
        raise_timeout: bool = False,
        # message kwargs
        headers: Optional[aio_pika.abc.HeadersType] = None,
        content_type: Optional[str] = None,
        content_encoding: Optional[str] = None,
        persist: bool = False,
        priority: Optional[int] = None,
        correlation_id: Optional[str] = None,
        reply_to: Optional[str] = None,
        expiration: Optional[aio_pika.abc.DateType] = None,
        message_id: Optional[str] = None,
        timestamp: Optional[aio_pika.abc.DateType] = None,
        type: Optional[str] = None,
        user_id: Optional[str] = None,
        app_id: Optional[str] = None,
    ) -> Optional[Union[aiormq.abc.ConfirmationFrameType, DecodedMessage]]:
        """Publish the message to the exchange with the routing key.

        Args:
            message: encodable message to send
            queue: if routing key is not set, use queue instead
            exchange: exchange to publish message. Use `default` if not specified
            routing_key: message routing key
            mandatory: wait for message will be placed in any queue
            immediate: expects available consumer
            timeout: request to RabbitMQ timeout
            headers: message headers (for consumers)
            content_type: message content-type to decode
            content_encoding: message encoding
            persist: restore message on RabbitMQ reboot
            priority: message priority
            correlation_id: correlation id to match message with response
            reply_to: queue to send response
            message_id: message identifier
            timestamp: message sending time
            expiration: message lifetime (in seconds)
            type: message type (for consumers)
            user_id: RabbitMQ user who sent the message
            app_id: application identifier (for consumers)
            callback: wait for response
            callback_timeout: response waiting time
            raise_timeout: if False timeout returns None instead TimeoutError

        Returns:
            `aiormq.abc.ConfirmationFrameType` if you are not waiting for response
            (reply_to and callback are not specified)

            `DecodedMessage` | `None` if response is expected

        _publisher confirms: https://www.rabbitmq.com/confirms.html
        """
    def handle(  # type: ignore[override]
        self,
        queue: Union[str, RabbitQueue],
        exchange: Union[str, RabbitExchange, None] = None,
        *,
        consume_arguments: Optional[AnyDict] = None,
        retry: Union[bool, int] = False,
        dependencies: Sequence[Depends] = (),
        decode_message: AsyncDecoder[IncomingMessage] = None,
        parse_message: AsyncParser[IncomingMessage] = None,
        # AsyncAPI
        description: str = "",
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[IncomingMessage, bool], Awaitable[T_HandlerReturn]],
    ]:
        """Register queue consumer method

        Args:
            queue: queue to consume messages
            exchange: exchange to bind queue
            retry: at message exception will returns to queue `int` times or endless if `True`
            dependencies: wrap handler dependencies
            decode_message: custom RabbitMessage decoder
            parse_message: custom IncomingMessage to RabbitMessage parser
            description: AsyncAPI channel object description

        Returns:
            Async or sync function decorator
        """
    async def start(self) -> None:
        """Initialize RabbitMQ connection and startup all consumers"""
    async def close(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_val: Optional[BaseException] = None,
        exec_tb: Optional[TracebackType] = None,
    ) -> None:
        """Close RabbitMQ connection"""
    async def declare_queue(self, queue: RabbitQueue) -> aio_pika.RobustQueue:
        """Check existence or create RabbitMQ queue"""
    async def declare_exchange(
        self, exchange: RabbitExchange
    ) -> aio_pika.RobustExchange:
        """Check existence or create RabbitMQ exchange"""
    @property
    def channel(self) -> aio_pika.RobustChannel:
        """Access to brokers' aio-pika channel object"""
    def _process_message(
        self,
        func: Callable[[RabbitMessage], Awaitable[T_HandlerReturn]],
        watcher: Optional[BaseWatcher],
    ) -> Callable[[RabbitMessage], Awaitable[T_HandlerReturn]]: ...
    def _get_log_context(  # type: ignore[override]
        self,
        message: Optional[RabbitMessage],
        queue: RabbitQueue,
        exchange: Optional[RabbitExchange] = None,
    ) -> Dict[str, Any]: ...
    async def _init_handler(
        self,
        handler: Handler,
    ) -> aio_pika.abc.AbstractRobustQueue: ...
    @classmethod
    def _validate_message(
        cls: Type["RabbitBroker"],
        message: PikaSendableMessage,
        callback_queue: Optional[aio_pika.abc.AbstractRobustQueue] = None,
        **message_kwargs: Dict[str, Any],
    ) -> aio_pika.Message: ...
    @staticmethod
    async def _parse_message(
        message: IncomingMessage,
    ) -> RabbitMessage: ...
    async def _connect(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> aio_pika.RobustConnection: ...



from typing import Optional

from pydantic import BaseModel

from propan.asyncapi.bindings.amqp import (
    AsyncAPIAmqpChannelBinding,
    AsyncAPIAmqpOperationBinding,
)
from propan.asyncapi.bindings.kafka import (
    AsyncAPIKafkaChannelBinding,
    AsyncAPIKafkaOperationBinding,
)
from propan.asyncapi.bindings.nats import (
    AsyncAPINatsChannelBinding,
    AsyncAPINatsOperationBinding,
)
from propan.asyncapi.bindings.redis import (
    AsyncAPIRedisChannelBinding,
    AsyncAPIRedisOperationBinding,
)
from propan.asyncapi.bindings.sqs import (
    AsyncAPISQSChannelBinding,
    AsyncAPISQSOperationBinding,
)


class AsyncAPIChannelBinding(BaseModel):
    amqp: Optional[AsyncAPIAmqpChannelBinding] = None
    kafka: Optional[AsyncAPIKafkaChannelBinding] = None
    sqs: Optional[AsyncAPISQSChannelBinding] = None
    nats: Optional[AsyncAPINatsChannelBinding] = None
    redis: Optional[AsyncAPIRedisChannelBinding] = None


class AsyncAPIOperationBinding(BaseModel):
    amqp: Optional[AsyncAPIAmqpOperationBinding] = None
    kafka: Optional[AsyncAPIKafkaOperationBinding] = None
    sqs: Optional[AsyncAPISQSOperationBinding] = None
    nats: Optional[AsyncAPINatsOperationBinding] = None
    redis: Optional[AsyncAPIRedisOperationBinding] = None



from nats.js.api import (
    ConsumerConfig,
    DeliverPolicy,
    DiscardPolicy,
    Placement,
    RePublish,
    RetentionPolicy,
    StorageType,
    StreamConfig,
    StreamSource,
)

__all__ = (
    "StreamConfig",
    "RetentionPolicy",
    "DiscardPolicy",
    "Placement",
    "StorageType",
    "StreamSource",
    "RePublish",
    "ConsumerConfig",
    "DeliverPolicy",
)



import logging
import ssl
from types import TracebackType
from typing import (
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

from fast_depends.dependencies import Depends
from nats.aio.client import (
    DEFAULT_CONNECT_TIMEOUT,
    DEFAULT_DRAIN_TIMEOUT,
    DEFAULT_INBOX_PREFIX,
    DEFAULT_MAX_FLUSHER_QUEUE_SIZE,
    DEFAULT_MAX_OUTSTANDING_PINGS,
    DEFAULT_MAX_RECONNECT_ATTEMPTS,
    DEFAULT_PENDING_SIZE,
    DEFAULT_PING_INTERVAL,
    DEFAULT_RECONNECT_TIME_WAIT,
    Callback,
    Client,
    Credentials,
    ErrorCallback,
    JWTCallback,
    SignatureCallback,
)
from nats.aio.msg import Msg
from nats.js.client import (
    DEFAULT_JS_SUB_PENDING_BYTES_LIMIT,
    DEFAULT_JS_SUB_PENDING_MSGS_LIMIT,
    JetStreamContext,
)

from propan.brokers._model.broker_usecase import (
    AsyncDecoder,
    AsyncParser,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers.middlewares import BaseMiddleware
from propan.brokers.nats import consts as api
from propan.brokers.nats.nats_broker import NatsBroker, NatsMessage
from propan.brokers.push_back_watcher import BaseWatcher
from propan.log import access_logger
from propan.types import DecodedMessage, SendableMessage

T = TypeVar("T")

class NatsJSBroker(NatsBroker):
    middlewares: Sequence[Type[BaseMiddleware[Msg]]]
    _raw_connection: Optional[Client]
    _connection: Optional[JetStreamContext]
    _stream_config: api.StreamConfig

    @property
    def js(self) -> JetStreamContext:
        """JetStreamContext object to use it with native `nats-py` features"""
    def __init__(
        self,
        servers: Union[str, List[str]] = ["nats://localhost:4222"],  # noqa: B006
        *,
        # JS
        stream: Optional[str] = "propan-stream",
        description: Optional[str] = None,
        retention: Optional[api.RetentionPolicy] = None,
        max_consumers: Optional[int] = None,
        max_msgs: Optional[int] = None,
        max_bytes: Optional[int] = None,
        discard: Optional[api.DiscardPolicy] = api.DiscardPolicy.OLD,
        max_age: Optional[float] = None,  # in seconds
        max_msgs_per_subject: int = -1,
        max_msg_size: Optional[int] = -1,
        storage: Optional[api.StorageType] = None,
        num_replicas: Optional[int] = None,
        no_ack: bool = False,
        template_owner: Optional[str] = None,
        duplicate_window: int = 0,
        placement: Optional[api.Placement] = None,
        mirror: Optional[api.StreamSource] = None,
        sources: Optional[List[api.StreamSource]] = None,
        sealed: bool = False,
        deny_delete: bool = False,
        deny_purge: bool = False,
        allow_rollup_hdrs: bool = False,
        republish: Optional[api.RePublish] = None,
        allow_direct: Optional[bool] = None,
        mirror_direct: Optional[bool] = None,
        # connection
        error_cb: Optional[ErrorCallback] = None,
        disconnected_cb: Optional[Callback] = None,
        closed_cb: Optional[Callback] = None,
        discovered_server_cb: Optional[Callback] = None,
        reconnected_cb: Optional[Callback] = None,
        name: Optional[str] = None,
        pedantic: bool = False,
        verbose: bool = False,
        allow_reconnect: bool = True,
        connect_timeout: int = DEFAULT_CONNECT_TIMEOUT,
        reconnect_time_wait: int = DEFAULT_RECONNECT_TIME_WAIT,
        max_reconnect_attempts: int = DEFAULT_MAX_RECONNECT_ATTEMPTS,
        ping_interval: int = DEFAULT_PING_INTERVAL,
        max_outstanding_pings: int = DEFAULT_MAX_OUTSTANDING_PINGS,
        dont_randomize: bool = False,
        flusher_queue_size: int = DEFAULT_MAX_FLUSHER_QUEUE_SIZE,
        no_echo: bool = False,
        tls: Optional[ssl.SSLContext] = None,
        tls_hostname: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
        drain_timeout: int = DEFAULT_DRAIN_TIMEOUT,
        signature_cb: Optional[SignatureCallback] = None,
        user_jwt_cb: Optional[JWTCallback] = None,
        user_credentials: Optional[Credentials] = None,
        nkeys_seed: Optional[str] = None,
        inbox_prefix: Union[str, bytes] = DEFAULT_INBOX_PREFIX,
        pending_size: int = DEFAULT_PENDING_SIZE,
        flush_timeout: Optional[float] = None,
        # broker kwargs
        logger: Optional[logging.Logger] = access_logger,
        log_level: int = logging.INFO,
        log_fmt: Optional[str] = None,
        apply_types: bool = True,
        dependencies: Sequence[Depends] = (),
        middlewares: Sequence[Type[BaseMiddleware[Msg]]] = (),
        decode_message: AsyncDecoder[Msg] = None,
        parse_message: AsyncParser[Msg] = None,
        protocol: str = "nats",
    ) -> None:
        """"""
    async def _connect(
        self,
        *,
        # connection
        servers: Union[str, List[str]] = ["nats://localhost:4222"],  # noqa: B006
        error_cb: Optional[ErrorCallback] = None,
        disconnected_cb: Optional[Callback] = None,
        closed_cb: Optional[Callback] = None,
        discovered_server_cb: Optional[Callback] = None,
        reconnected_cb: Optional[Callback] = None,
        name: Optional[str] = None,
        pedantic: bool = False,
        verbose: bool = False,
        allow_reconnect: bool = True,
        connect_timeout: int = DEFAULT_CONNECT_TIMEOUT,
        reconnect_time_wait: int = DEFAULT_RECONNECT_TIME_WAIT,
        max_reconnect_attempts: int = DEFAULT_MAX_RECONNECT_ATTEMPTS,
        ping_interval: int = DEFAULT_PING_INTERVAL,
        max_outstanding_pings: int = DEFAULT_MAX_OUTSTANDING_PINGS,
        dont_randomize: bool = False,
        flusher_queue_size: int = DEFAULT_MAX_FLUSHER_QUEUE_SIZE,
        no_echo: bool = False,
        tls: Optional[ssl.SSLContext] = None,
        tls_hostname: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
        drain_timeout: int = DEFAULT_DRAIN_TIMEOUT,
        signature_cb: Optional[SignatureCallback] = None,
        user_jwt_cb: Optional[JWTCallback] = None,
        user_credentials: Optional[Credentials] = None,
        nkeys_seed: Optional[str] = None,
        inbox_prefix: Union[str, bytes] = DEFAULT_INBOX_PREFIX,
        pending_size: int = DEFAULT_PENDING_SIZE,
        flush_timeout: Optional[float] = None,
    ) -> JetStreamContext:
        """"""
    async def close(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_val: Optional[BaseException] = None,
        exec_tb: Optional[TracebackType] = None,
    ) -> None:
        """"""
    async def _start(self):
        """"""
    def _process_message(
        self,
        func: Callable[[NatsMessage], Awaitable[T]],
        watcher: Optional[BaseWatcher] = None,
    ) -> Callable[[NatsMessage], Awaitable[T]]:
        """"""
    async def _parse_message(self, message: Msg) -> NatsMessage:
        """"""
    def handle(
        self,
        subject: str,
        queue: str = "",
        *,
        durable: Optional[str] = None,
        config: Optional[api.ConsumerConfig] = None,
        ordered_consumer: bool = False,
        idle_heartbeat: Optional[float] = None,
        flow_control: bool = False,
        pending_msgs_limit: Optional[int] = DEFAULT_JS_SUB_PENDING_MSGS_LIMIT,
        pending_bytes_limit: Optional[int] = DEFAULT_JS_SUB_PENDING_BYTES_LIMIT,
        deliver_policy: Optional[api.DeliverPolicy] = None,
        headers_only: Optional[bool] = None,
        # broker kwargs
        retry: Union[bool, int] = False,
        dependencies: Sequence[Depends] = (),
        decode_message: AsyncDecoder[Msg] = None,
        parse_message: AsyncParser[Msg] = None,
        description: str = "",
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[Msg, bool], Awaitable[T_HandlerReturn]],
    ]:
        """"""
    async def publish(
        self,
        message: SendableMessage,
        subject: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        reply_to: str = "",
        callback: bool = False,
        callback_timeout: Optional[float] = 30.0,
        raise_timeout: bool = False,
    ) -> Optional[DecodedMessage]:
        """"""



import asyncio
from dataclasses import dataclass
from typing import Any, Dict, Optional

from fast_depends.core import CallModel
from pydantic import BaseModel, Field
from redis.asyncio.client import PubSub

from propan.asyncapi.bindings import (
    AsyncAPIChannelBinding,
    AsyncAPIOperationBinding,
    redis,
)
from propan.asyncapi.channels import AsyncAPIChannel
from propan.asyncapi.message import AsyncAPIMessage
from propan.asyncapi.subscription import AsyncAPISubscription
from propan.brokers._model.schemas import BaseHandler
from propan.types import DecoratedCallable


@dataclass
class Handler(BaseHandler):
    channel: str
    pattern: bool = False

    task: Optional["asyncio.Task[Any]"] = None
    subscription: Optional[PubSub] = None

    def __init__(
        self,
        callback: DecoratedCallable,
        dependant: CallModel,
        channel: str,
        pattern: bool = False,
        task: Optional["asyncio.Task[Any]"] = None,
        subscription: Optional[PubSub] = None,
        _description: str = "",
    ):
        self.callback = callback
        self.dependant = dependant
        self._description = _description
        self.channel = channel
        self.pattern = pattern
        self.task = task
        self.subscription = subscription

    def get_schema(self) -> Dict[str, AsyncAPIChannel]:
        message_title, body, reply_to = self.get_message_object()

        return {
            self.title: AsyncAPIChannel(
                subscribe=AsyncAPISubscription(
                    description=self.description,
                    bindings=AsyncAPIOperationBinding(
                        redis=redis.AsyncAPIRedisOperationBinding(
                            replyTo=reply_to,
                        ),
                    ),
                    message=AsyncAPIMessage(
                        title=message_title,
                        payload=body,
                    ),
                ),
                bindings=AsyncAPIChannelBinding(
                    redis=redis.AsyncAPIRedisChannelBinding(
                        channel=self.channel,
                        method="psubscribe" if self.pattern else "subscribe",
                    )
                ),
            ),
        }


class RedisMessage(BaseModel):
    data: bytes
    headers: Dict[str, str] = Field(default_factory=dict)
    reply_to: str = ""



from propan import PropanApp, RedisBroker
from propan.annotations import Logger

broker = RedisBroker()
app = PropanApp(broker)


@broker.handle("*.info", pattern=True)
async def handler1(b: str, logger: Logger):
    logger.info("handler1")


@broker.handle("*.info", pattern=True)
async def handler2(b: str, logger: Logger):
    logger.info("handler2")


@broker.handle("*.error", pattern=True)
async def handler3(logger: Logger):
    logger.info("handler3")


@app.after_startup
async def publish_smth():
    await broker.publish("", "logs.info")
    await broker.publish("", "logs.error")



from propan.brokers._model.broker_usecase import BrokerAsyncUsecase
from propan.brokers._model.routing import BrokerRouter
from propan.brokers._model.schemas import PropanMessage, Queue

__all__ = (
    "Queue",
    "BrokerAsyncUsecase",
    "PropanMessage",
    "BrokerRouter",
)



import logging
from types import TracebackType
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

from fast_depends.dependencies import Depends
from redis.asyncio.client import Redis
from redis.asyncio.connection import BaseParser, Connection, DefaultParser, Encoder
from typing_extensions import TypeAlias

from propan.brokers._model import BrokerAsyncUsecase
from propan.brokers._model.broker_usecase import (
    AsyncDecoder,
    AsyncParser,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers._model.schemas import PropanMessage
from propan.brokers.middlewares import BaseMiddleware
from propan.brokers.push_back_watcher import BaseWatcher
from propan.brokers.redis.schemas import Handler
from propan.log import access_logger
from propan.types import AnyDict, DecodedMessage, SendableMessage

T = TypeVar("T")
RedisMessage: TypeAlias = PropanMessage[AnyDict]

class RedisBroker(BrokerAsyncUsecase[AnyDict, Redis[bytes]]):
    handlers: List[Handler]
    middlewares: Sequence[Type[BaseMiddleware[AnyDict]]]
    __max_channel_len: int

    def __init__(
        self,
        url: str = "redis://localhost:6379",
        *,
        polling_interval: float = 1.0,
        host: str = "localhost",
        port: Union[str, int] = 6379,
        username: Optional[str] = None,
        password: Optional[str] = None,
        db: Union[str, int] = 0,
        client_name: Optional[str] = None,
        health_check_interval: float = 0,
        max_connections: Optional[int] = None,
        socket_timeout: Optional[float] = None,
        socket_connect_timeout: Optional[float] = None,
        socket_read_size: int = 65536,
        socket_keepalive: bool = False,
        socket_keepalive_options: Optional[Mapping[int, Union[int, bytes]]] = None,
        socket_type: int = 0,
        retry_on_timeout: bool = False,
        encoding: str = "utf-8",
        encoding_errors: str = "strict",
        decode_responses: bool = False,
        parser_class: Type[BaseParser] = DefaultParser,
        connection_class: Type[Connection] = Connection,
        encoder_class: Type[Encoder] = Encoder,
        # broker kwargs
        logger: Optional[logging.Logger] = access_logger,
        log_level: int = logging.INFO,
        log_fmt: Optional[str] = None,
        apply_types: bool = True,
        dependencies: Sequence[Depends] = (),
        middlewares: Sequence[Type[BaseMiddleware[AnyDict]]] = (),
        decode_message: AsyncDecoder[AnyDict] = None,
        parse_message: AsyncParser[AnyDict] = None,
        protocol: str = "redis",
    ) -> None:
        """Redis Pub/sub Propan broker

        URL examples:

        - redis://[[username]:[password]]@localhost:6379/0
        - rediss://[[username]:[password]]@localhost:6379/0
        - unix://[username@]/path/to/socket.sock?db=0[&password=password]

        Three URL schemes are supported:

        - `redis://` creates a TCP socket connection. See more at:
          <https://www.iana.org/assignments/uri-schemes/prov/redis>
        - `rediss://` creates a SSL wrapped TCP socket connection. See more at:
          <https://www.iana.org/assignments/uri-schemes/prov/rediss>
        - `unix://`: creates a Unix Domain Socket connection.

        Url will be parsed to kwargs and partially replaced by keywords arguments if they specified.
        """
    async def connect(
        self,
        *,
        url: str = "redis://localhost:6379",
        host: str = "localhost",
        port: Union[str, int] = 6379,
        username: Optional[str] = None,
        password: Optional[str] = None,
        db: Union[str, int] = 0,
        client_name: Optional[str] = None,
        health_check_interval: float = 0,
        max_connections: Optional[int] = None,
        socket_timeout: Optional[float] = None,
        socket_connect_timeout: Optional[float] = None,
        socket_read_size: int = 65536,
        socket_keepalive: bool = False,
        socket_keepalive_options: Optional[Mapping[int, Union[int, bytes]]] = None,
        socket_type: int = 0,
        retry_on_timeout: bool = False,
        encoding: str = "utf-8",
        encoding_errors: str = "strict",
        decode_responses: bool = False,
        parser_class: Type[BaseParser] = DefaultParser,
        connection_class: Type[Connection] = Connection,
        encoder_class: Type[Encoder] = Encoder,
    ) -> Redis[bytes]:
        """Connect to Redis

        URL examples:

        - redis://[[username]:[password]]@localhost:6379/0
        - rediss://[[username]:[password]]@localhost:6379/0
        - unix://[username@]/path/to/socket.sock?db=0[&password=password]

        Three URL schemes are supported:

        - `redis://` creates a TCP socket connection. See more at:
          <https://www.iana.org/assignments/uri-schemes/prov/redis>
        - `rediss://` creates a SSL wrapped TCP socket connection. See more at:
          <https://www.iana.org/assignments/uri-schemes/prov/rediss>
        - `unix://`: creates a Unix Domain Socket connection.

        Url will be parsed to kwargs and partially replaced by keywords arguments if they specified.
        """
    async def _connect(self, *args: Any, **kwargs: Any) -> Redis[bytes]: ...
    async def start(self) -> None:
        """Initialize Redis connection and startup all consumers"""
    async def close(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_val: Optional[BaseException] = None,
        exec_tb: Optional[TracebackType] = None,
    ) -> None:
        """Cancel all consumers tasks and subscribtions, close Redis connection"""
    def handle(  # type: ignore[override]
        self,
        channel: str,
        *,
        pattern: bool = False,
        dependencies: Sequence[Depends] = (),
        decode_message: AsyncDecoder[AnyDict] = None,
        parse_message: AsyncParser[AnyDict] = None,
        description: str = "",
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[AnyDict, bool], Awaitable[T_HandlerReturn]],
    ]:
        """Register channel consumer method

        Args:
            channel: channel to consume messages
            pattern: use psubscribe or subscribe method
            dependencies: wrap handler dependencies
            decode_message: custom PropanMessage[AnyDict] decoder
            parse_message: custom redis message to PropanMessage[AnyDict] parser
            description: AsyncAPI channel object description

        Returns:
            Async or sync function decorator
        """
    async def publish(  # type: ignore[override]
        self,
        message: SendableMessage = "",
        channel: str = "",
        *,
        reply_to: str = "",
        headers: Optional[Dict[str, Any]] = None,
        callback: bool = False,
        callback_timeout: Optional[float] = 30.0,
        raise_timeout: bool = False,
    ) -> Optional[DecodedMessage]:
        """Publish the message to the channel.

        Args:
            message: encodable message to send
            channel: channel to publish message
            reply_to: queue to send response
            headers: message headers (for consumers)
            callback: wait for response
            callback_timeout: response waiting time
            raise_timeout: if False timeout returns None instead TimeoutError

        Returns:
            `None` if you are not waiting for response
            (reply_to and callback are not specified)

            `DecodedMessage` | `None` if response is expected
        """
    def _get_log_context(  # type: ignore[override]
        self, message: Optional[RedisMessage], channel: str
    ) -> Dict[str, Any]: ...
    @staticmethod
    async def _decode_message(message: RedisMessage) -> DecodedMessage: ...
    @staticmethod
    async def _parse_message(message: AnyDict) -> RedisMessage: ...
    def _process_message(
        self,
        func: Callable[[RedisMessage], Awaitable[T]],
        watcher: Optional[BaseWatcher],
    ) -> Callable[[RedisMessage], Awaitable[T]]: ...



import logging
from asyncio import AbstractEventLoop, Future
from ssl import SSLContext
from types import TracebackType
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

from aiokafka import AIOKafkaConsumer, AIOKafkaProducer
from aiokafka.abc import AbstractTokenProvider
from aiokafka.producer.producer import _missing
from aiokafka.structs import ConsumerRecord
from fast_depends.dependencies import Depends
from kafka.coordinator.assignors.abstract import AbstractPartitionAssignor
from kafka.coordinator.assignors.roundrobin import RoundRobinPartitionAssignor
from kafka.partitioner.default import DefaultPartitioner
from typing_extensions import Literal, TypeAlias, TypeVar

from propan.__about__ import __version__
from propan.brokers._model.broker_usecase import (
    AsyncDecoder,
    AsyncParser,
    BrokerAsyncUsecase,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers._model.schemas import PropanMessage
from propan.brokers.kafka.schemas import Handler
from propan.brokers.middlewares import BaseMiddleware
from propan.brokers.push_back_watcher import BaseWatcher
from propan.log import access_logger
from propan.types import DecodedMessage, SendableMessage

Partition = TypeVar("Partition")
CorrelationId: TypeAlias = str
KafkaMessage: TypeAlias = PropanMessage[ConsumerRecord]

class KafkaBroker(
    BrokerAsyncUsecase[ConsumerRecord, Callable[[Tuple[str, ...]], AIOKafkaConsumer]]
):
    _publisher: Optional[AIOKafkaProducer]
    __max_topic_len: int
    response_topic: str
    response_callbacks: Dict[CorrelationId, "Future[DecodedMessage]"]
    handlers: List[Handler]
    middlewares: Sequence[Type[BaseMiddleware[ConsumerRecord]]]

    def __init__(
        self,
        bootstrap_servers: Union[str, List[str]] = "localhost",
        *,
        response_topic: str = "",
        # both
        client_id: str = "propan-" + __version__,
        request_timeout_ms: int = 40 * 1000,
        retry_backoff_ms: int = 100,
        metadata_max_age_ms: int = 5 * 60 * 1000,
        security_protocol: Literal[
            "SSL",
            "PLAINTEXT",
        ] = "PLAINTEXT",
        api_version: str = "auto",
        connections_max_idle_ms: int = 540000,
        sasl_mechanism: Literal[
            "PLAIN",
            "GSSAPI",
            "SCRAM-SHA-256",
            "SCRAM-SHA-512",
            "OAUTHBEARER",
        ] = "PLAIN",
        sasl_plain_password: Optional[str] = None,
        sasl_plain_username: Optional[str] = None,
        sasl_kerberos_service_name: str = "kafka",
        sasl_kerberos_domain_name: Optional[str] = None,
        sasl_oauth_token_provider: Optional[AbstractTokenProvider] = None,
        # publisher
        acks: Union[Literal[0, 1, -1, "all"], object] = _missing,
        key_serializer: Optional[Callable[[Any], bytes]] = None,
        value_serializer: Optional[Callable[[Any], bytes]] = None,
        compression_type: Optional[Literal["gzip", "snappy", "lz4", "zstd"]] = None,
        max_batch_size: int = 16384,
        partitioner: Callable[
            [bytes, List[Partition], List[Partition]],
            Partition,
        ] = DefaultPartitioner(),
        max_request_size: int = 1048576,
        linger_ms: int = 0,
        send_backoff_ms: int = 100,
        ssl_context: Optional[SSLContext] = None,
        enable_idempotence: bool = False,
        transactional_id: Optional[str] = None,
        transaction_timeout_ms: int = 60000,
        loop: Optional[AbstractEventLoop] = None,
        # broker
        logger: Optional[logging.Logger] = access_logger,
        decode_message: AsyncDecoder[ConsumerRecord] = None,
        parse_message: AsyncParser[ConsumerRecord] = None,
        log_level: int = logging.INFO,
        log_fmt: Optional[str] = None,
        apply_types: bool = True,
        dependencies: Sequence[Depends] = (),
        middlewares: Sequence[Type[BaseMiddleware[ConsumerRecord]]] = (),
        protocol: str = "kafka",
    ) -> None: ...
    async def connect(
        self,
        *,
        bootstrap_servers: Union[str, List[str]] = "localhost",
        # both
        loop: Optional[AbstractEventLoop] = None,
        client_id: str = "propan-" + __version__,
        request_timeout_ms: int = 40 * 1000,
        retry_backoff_ms: int = 100,
        metadata_max_age_ms: int = 5 * 60 * 1000,
        security_protocol: Literal[
            "SSL",
            "PLAINTEXT",
        ] = "PLAINTEXT",
        api_version: str = "auto",
        connections_max_idle_ms: int = 540000,
        sasl_mechanism: Literal[
            "PLAIN",
            "GSSAPI",
            "SCRAM-SHA-256",
            "SCRAM-SHA-512",
            "OAUTHBEARER",
        ] = "PLAIN",
        sasl_plain_password: Optional[str] = None,
        sasl_plain_username: Optional[str] = None,
        sasl_kerberos_service_name: str = "kafka",
        sasl_kerberos_domain_name: Optional[str] = None,
        sasl_oauth_token_provider: Optional[AbstractTokenProvider] = None,
        # publisher
        acks: Union[Literal[0, 1, -1, "all"], object] = _missing,
        key_serializer: Optional[Callable[[Any], bytes]] = None,
        value_serializer: Optional[Callable[[Any], bytes]] = None,
        compression_type: Optional[Literal["gzip", "snappy", "lz4", "zstd"]] = None,
        max_batch_size: int = 16384,
        partitioner: Callable[
            [bytes, List[Partition], List[Partition]],
            Partition,
        ] = DefaultPartitioner(),
        max_request_size: int = 1048576,
        linger_ms: int = 0,
        send_backoff_ms: int = 100,
        ssl_context: Optional[SSLContext] = None,
        enable_idempotence: bool = False,
        transactional_id: Optional[str] = None,
        transaction_timeout_ms: int = 60000,
    ) -> AIOKafkaConsumer: ...
    async def _connect(self, *args: Any, **kwargs: Any) -> AIOKafkaConsumer: ...
    async def close(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_val: Optional[BaseException] = None,
        exec_tb: Optional[TracebackType] = None,
    ) -> None: ...
    def handle(  # type: ignore[override]
        self,
        *topics: str,
        group_id: Optional[str] = None,
        key_deserializer: Optional[Callable[[bytes], Any]] = None,
        value_deserializer: Optional[Callable[[bytes], Any]] = None,
        fetch_max_wait_ms: int = 500,
        fetch_max_bytes: int = 52428800,
        fetch_min_bytes: int = 1,
        max_partition_fetch_bytes: int = 1 * 1024 * 1024,
        auto_offset_reset: Literal[
            "latest",
            "earliest",
            "none",
        ] = "latest",
        enable_auto_commit: bool = True,
        auto_commit_interval_ms: int = 5000,
        check_crcs: bool = True,
        partition_assignment_strategy: Sequence[AbstractPartitionAssignor] = (
            RoundRobinPartitionAssignor,
        ),
        max_poll_interval_ms: int = 300000,
        rebalance_timeout_ms: Optional[int] = None,
        session_timeout_ms: int = 10000,
        heartbeat_interval_ms: int = 3000,
        consumer_timeout_ms: int = 200,
        max_poll_records: Optional[int] = None,
        exclude_internal_topics: bool = True,
        isolation_level: Literal[
            "read_uncommitted",
            "read_committed",
        ] = "read_uncommitted",
        retry: Union[bool, int] = False,
        dependencies: Sequence[Depends] = (),
        decode_message: AsyncDecoder[ConsumerRecord] = None,
        parse_message: AsyncParser[ConsumerRecord] = None,
        description: str = "",
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[ConsumerRecord, bool], Awaitable[T_HandlerReturn]],
    ]: ...
    @staticmethod
    async def _parse_message(
        message: ConsumerRecord,
    ) -> KafkaMessage: ...
    async def publish(  # type: ignore[override]
        self,
        message: SendableMessage,
        topic: str,
        key: Optional[bytes] = None,
        partition: Optional[int] = None,
        timestamp_ms: Optional[int] = None,
        headers: Optional[Dict[str, str]] = None,
        *,
        reply_to: str = "",
        callback: bool = False,
        callback_timeout: Optional[float] = None,
        raise_timeout: bool = False,
    ) -> Optional[DecodedMessage]: ...
    async def publish_batch(
        self,
        *msgs: SendableMessage,
        topic: str,
        partition: Optional[int] = None,
        timestamp_ms: Optional[int] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> None: ...
    def _get_log_context(  # type: ignore[override]
        self,
        message: Optional[KafkaMessage],
        topics: Sequence[str] = (),
    ) -> Dict[str, Any]: ...
    def _process_message(
        self,
        func: Callable[[KafkaMessage], Awaitable[T_HandlerReturn]],
        watcher: Optional[BaseWatcher],
    ) -> Callable[[KafkaMessage], Awaitable[T_HandlerReturn]]: ...



from inspect import _empty
from typing import Any

from fast_depends.library import CustomField

from propan.types import AnyDict
from propan.utils.context import context


class Context(CustomField):  # type: ignore
    def __init__(
        self,
        real_name: str = "",
        *,
        cast: bool = False,
        default: Any = _empty,
    ):
        self.name = real_name
        self.default = default
        super().__init__(
            cast=cast,
            required=(default is _empty),
        )

    def use(self, **kwargs: AnyDict) -> AnyDict:
        name = self.name or self.param_name

        try:
            kwargs[self.param_name] = resolve_context(name)
        except (KeyError, AttributeError):
            if self.required is False:
                kwargs[self.param_name] = self.default

        return kwargs


def resolve_context(argument: str) -> Any:
    keys = argument.split(".")

    v = context.context[keys[0]]
    for i in keys[1:]:
        v = getattr(v, i)

    return v



import sys
from pathlib import Path

import typer

from propan.cli.docs.gen import (
    generate_doc_file,
    get_app_schema,
    json_schema_to_yaml,
    schema_to_json,
)
from propan.cli.docs.serving import serve_docs
from propan.cli.utils.imports import get_app_path, try_import_propan

docs_app = typer.Typer(pretty_exceptions_short=True)


@docs_app.command(name="gen")
def gen(
    app: str = typer.Argument(
        ...,
        help="[python_module:PropanApp] - path to your application",
    ),
    filename: str = typer.Option(
        "asyncapi.yaml",
        "-f",
        "--f",
        case_sensitive=False,
        show_default=True,
        help="generated document filename",
    ),
) -> None:
    """Generate an AsyncAPI scheme.yaml for your project"""
    current_dir = Path.cwd()
    generated_filepath = current_dir / filename

    module, app = get_app_path(app)
    app_dir = module.parent
    sys.path.insert(0, str(app_dir))
    propan_app = try_import_propan(module, app)

    generate_doc_file(propan_app, generated_filepath)


@docs_app.command(name="serve")
def serve(
    app: str = typer.Argument(
        ...,
        help="[python_module:PropanApp] or [asyncapi.yaml] - path to your application documentation",
    ),
    host: str = typer.Option(
        "localhost",
        help="documentation hosting address",
    ),
    port: int = typer.Option(
        8000,
        help="documentation hosting port",
    ),
) -> None:
    """Serve project AsyncAPI scheme"""
    if ":" in app:
        module, app = get_app_path(app)
        app_dir = module.parent
        sys.path.insert(0, str(app_dir))
        propan_app = try_import_propan(module, app)
        raw_schema = get_app_schema(propan_app)
        schema = json_schema_to_yaml(schema_to_json(raw_schema))

    else:
        schema_filepath = Path.cwd() / app
        schema = schema_filepath.read_text()
        raw_schema = None

    serve_docs(
        schema=schema,
        host=host,
        port=port,
        raw_schema=raw_schema,
    )



import asyncio
import logging
from functools import wraps
from types import TracebackType
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    NoReturn,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)
from uuid import uuid4

import anyio
from aiobotocore.client import AioBaseClient
from aiobotocore.session import get_session
from fast_depends.dependencies import Depends
from typing_extensions import TypeAlias

from propan._compat import model_to_dict
from propan.brokers._model.broker_usecase import (
    BrokerAsyncUsecase,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers._model.schemas import PropanMessage
from propan.brokers.exceptions import SkipMessage
from propan.brokers.push_back_watcher import (
    BaseWatcher,
    NotPushBackWatcher,
    WatcherContext,
)
from propan.brokers.sqs.schema import Handler
from propan.brokers.sqs.schema import SQSMessage as SM
from propan.brokers.sqs.schema import SQSQueue
from propan.types import AnyDict, DecodedMessage, SendableMessage
from propan.utils import context

T = TypeVar("T")
QueueUrl: TypeAlias = str
CorrelationId: TypeAlias = str
SQSMessage: TypeAlias = PropanMessage[AnyDict]


class SQSBroker(BrokerAsyncUsecase[AnyDict, AioBaseClient]):
    _connection: AioBaseClient
    _queues: Dict[str, QueueUrl]
    __max_queue_len: int
    response_queue: str
    response_callbacks: Dict[CorrelationId, "asyncio.Future[DecodedMessage]"]
    handlers: List[Handler]

    def __init__(
        self,
        url: str = "http://localhost:9324/",
        *,
        log_fmt: Optional[str] = None,
        response_queue: str = "",
        protocol: str = "sqs",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            url,
            log_fmt=log_fmt,
            url_=url,
            protocol=protocol,
            **kwargs,
        )
        self._queues = {}
        self.__max_queue_len = 4
        self.response_queue = response_queue
        self.response_callbacks = {}

    async def _connect(self, *, url: str, **kwargs: Any) -> AioBaseClient:
        session = get_session()
        client: AioBaseClient = await session._create_client(
            service_name="sqs", endpoint_url=url, **kwargs
        )
        context.set_global("client", client)
        await client.__aenter__()
        return client

    async def close(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc_val: Optional[BaseException] = None,
        exec_tb: Optional[TracebackType] = None,
    ) -> None:
        await super().close(exc_type, exc_val, exec_tb)
        for f in self.response_callbacks.values():
            f.cancel()
        self.response_callbacks = {}

        for h in self.handlers:
            if h.task is not None:
                h.task.cancel()
                h.task = None

        if self._connection is not None:
            await self._connection.__aexit__(None, None, None)
            self._connection = None

    async def _parse_message(self, message: Dict[str, Any]) -> SQSMessage:
        attributes = message.get("MessageAttributes", {})

        headers = {i: j.get("StringValue") for i, j in attributes.items()}

        return SQSMessage(
            body=message.get("Body", "").encode(),
            message_id=message.get("MessageId"),
            content_type=headers.pop("content-type", None),
            reply_to=headers.pop("reply_to", None) or "",
            headers=headers,
            raw_message=message,
        )

    def _process_message(
        self,
        func: Callable[[SQSMessage], Awaitable[T]],
        watcher: Optional[BaseWatcher],
    ) -> Callable[[SQSMessage], Awaitable[T]]:
        if watcher is None:
            watcher = NotPushBackWatcher()

        @wraps(func)
        async def process_wrapper(message: SQSMessage) -> T:
            context = WatcherContext(
                watcher,
                message,
                on_success=delete_message,
                on_max=delete_message,
                connection=self._connection,
            )

            async with context:
                r = await func(message)

                if message.reply_to:
                    await self.publish(
                        message=r or "",
                        queue=message.reply_to,
                        headers={
                            "correlation_id": message.headers.get("correlation_id")
                        },
                    )

                return r

        return process_wrapper

    def handle(
        self,
        queue: Union[str, SQSQueue],
        *,
        wait_interval: int = 1,
        max_messages_number: int = 10,  # 1...10
        attributes: Sequence[str] = (),
        message_attributes: Sequence[str] = (),
        request_attempt_id: Optional[str] = None,
        visibility_timeout: int = 0,
        dependencies: Sequence[Depends] = (),
        description: str = "",
        **original_kwargs: AnyDict,
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[Any, bool], Awaitable[T_HandlerReturn]],
    ]:
        super().handle()

        if isinstance(queue, str):
            queue = SQSQueue(queue)

        self.__max_queue_len = max((self.__max_queue_len, len(queue.name)))

        params = {
            "WaitTimeSeconds": wait_interval,
            "MaxNumberOfMessages": max_messages_number,
            "AttributeNames": [*attributes],
            "VisibilityTimeout": visibility_timeout,
            "MessageAttributeNames": (
                "content-type",
                "reply_to",
                "correlation_id",
                *message_attributes,
            ),
        }
        if request_attempt_id is not None:
            params["ReceiveRequestAttemptId"] = request_attempt_id

        def wrapper(func: HandlerCallable) -> HandlerCallable:
            func, dependant = self._wrap_handler(
                func,
                queue=queue.name,
                extra_dependencies=dependencies,
                **original_kwargs,
            )
            handler = Handler(
                callback=func,
                queue=queue,
                consumer_params=params,
                _description=description,
                dependant=dependant,
            )
            self.handlers.append(handler)
            return func

        return wrapper

    async def start(self) -> None:
        if self.response_queue:
            self.handle(
                self.response_queue,
                _raw=True,
            )(self._consume_response)

        context.set_local(
            "log_context",
            self._get_log_context(None, ""),
        )

        await super().start()

        for handler in self.handlers:  # pragma: no branch
            c = self._get_log_context(None, handler.queue.name)
            self._log(f"`{handler.callback.__name__}` waiting for messages", extra=c)

            url = await self.create_queue(handler.queue)
            handler.task = asyncio.create_task(self._consume(url, handler))

    async def publish(
        self,
        message: SendableMessage,
        queue: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        delay_seconds: int = 0,  # 0...900
        message_attributes: Optional[Dict[str, Any]] = None,
        message_system_attributes: Optional[Dict[str, Any]] = None,
        # FIFO only
        deduplication_id: Optional[str] = None,
        group_id: Optional[str] = None,
        # broker
        reply_to: str = "",
        callback: bool = False,
        callback_timeout: Optional[float] = None,
        raise_timeout: bool = False,
    ) -> None:
        queue_url = await self.get_queue(queue)

        if callback is True:
            reply_to = reply_to or self.response_queue
            if not reply_to:
                raise ValueError(
                    "You should specify `response_queue` at init or use `reply_to` argument"
                )

        if reply_to:
            correlation_id = str(uuid4())
        else:
            correlation_id = ""

        response_future: Optional["asyncio.Future[DecodedMessage]"]
        if callback is True:
            response_future = asyncio.Future()
            self.response_callbacks[correlation_id] = response_future
        else:
            response_future = None

        params = SM(
            message=message,
            headers=headers or {},
            delay_seconds=delay_seconds,
            message_attributes=message_attributes or {},
            message_system_attributes=message_system_attributes or {},
            deduplication_id=deduplication_id,
            group_id=group_id,
        ).to_params(
            reply_to=reply_to,
            correlation_id=correlation_id,
        )

        await self._connection.send_message(QueueUrl=queue_url, **params)

        if response_future is not None:
            if raise_timeout:
                scope = anyio.fail_after
            else:
                scope = anyio.move_on_after

            msg: Any = None
            with scope(callback_timeout):
                msg = await response_future

            if msg:
                return await self._decode_message(msg)

    async def create_queue(self, queue: SQSQueue) -> QueueUrl:
        url = self._queues.get(queue.name)
        if url is None:  # pragma: no branch
            url = (
                await self._connection.create_queue(
                    QueueName=queue.name,
                    Attributes={
                        i: str(j)
                        for i, j in model_to_dict(
                            queue,
                            exclude={"name", "tags"},
                            by_alias=True,
                            exclude_defaults=True,
                            exclude_unset=True,
                        ).items()
                    },
                    tags=queue.tags,
                )
            ).get("QueueUrl", "")
            self._queues[queue.name] = url
        return url

    async def delete_queue(self, queue: str) -> None:
        await self._connection.delete_queue(QueueUrl=self.get_queue(queue))
        self._queues.pop(queue)

    async def get_queue(self, queue: str) -> QueueUrl:
        url = self._queues.get(queue)
        if url is None:
            url = (await self._connection.get_queue_url(QueueName=queue)).get(
                "QueueUrl", ""
            )
            self._queues[queue] = url
        return url

    async def _consume(self, queue_url: str, handler: Handler) -> NoReturn:
        c = self._get_log_context(None, handler.queue.name)

        connected = True
        with context.scope("queue_url", queue_url):
            while True:
                try:
                    if connected is False:
                        await self.create_queue(handler.queue)

                    r = await self._connection.receive_message(
                        QueueUrl=queue_url,
                        **handler.consumer_params,
                    )

                except Exception as e:
                    if connected is True:
                        self._log(e, logging.WARNING, c)
                        self._queues.pop(handler.queue.name)
                        connected = False

                    await anyio.sleep(5)

                else:
                    if connected is False:
                        self._log("Connection established", logging.INFO, c)
                        connected = True

                    messages = r.get("Messages", [])
                    for msg in messages:
                        try:
                            await handler.callback(msg, True)
                        except Exception:
                            has_trash_messages = True
                        else:
                            has_trash_messages = False

                    if has_trash_messages is True:
                        await anyio.sleep(
                            handler.consumer_params.get("WaitTimeSeconds", 1.0)
                        )

    async def _consume_response(self, message: SQSMessage):
        correlation_id = message.headers.get("correlation_id")
        if correlation_id is not None:
            callback = self.response_callbacks.pop(correlation_id, None)
            if callback is not None:
                callback.set_result(message)
                return

        raise SkipMessage()

    @property
    def fmt(self) -> str:
        return self._fmt or (
            "%(asctime)s %(levelname)s - "
            f"%(queue)-{self.__max_queue_len}s | "
            "%(message_id)-10s "
            "- %(message)s"
        )

    def _get_log_context(
        self, message: Optional[SQSMessage], queue: str
    ) -> Dict[str, Any]:
        context = {
            "queue": queue,
            **super()._get_log_context(message),
        }
        return context


async def delete_message(
    message: SQSMessage, connection: Optional[AioBaseClient]
) -> None:
    if connection:
        await connection.delete_message(
            QueueUrl=context.get_local("queue_url"),
            ReceiptHandle=message.raw_message.get("ReceiptHandle", ""),
        )



from typing import Optional

from pydantic import BaseModel, Field, HttpUrl

from propan._compat import is_installed

if is_installed("email_validator"):
    from pydantic import EmailStr
else:  # pragma: no cover
    EmailStr = str  # type: ignore


class AsyncAPIContact(BaseModel):
    name: str
    url: HttpUrl
    email: Optional[EmailStr] = None


class AsyncAPILicense(BaseModel):
    name: str
    url: HttpUrl


class AsyncAPIInfo(BaseModel):
    title: str
    version: str = "1.0.0"
    description: str = ""
    terms: Optional[HttpUrl] = Field(
        default=None,
        alias="termsOfService",
    )
    contact: Optional[AsyncAPIContact] = None
    license: Optional[AsyncAPILicense] = None



from datetime import datetime
from typing import Any, Awaitable, Callable, Dict, Sequence, Union

from pydantic import BaseModel
from typing_extensions import TypeAlias

AsyncFunc: TypeAlias = Callable[..., Awaitable[Any]]

AnyDict: TypeAlias = Dict[str, Any]
AnyCallable: TypeAlias = Callable[..., Any]
NoneCallable: TypeAlias = Callable[..., None]

DecoratedCallable: TypeAlias = AnyCallable
DecoratedCallableNone: TypeAlias = NoneCallable

JsonDecodable: TypeAlias = Union[
    float,
    int,
    bool,
    str,
    bytes,
]
DecodedMessage: TypeAlias = Union[
    Dict[str, JsonDecodable], Sequence[JsonDecodable], JsonDecodable
]
SendableMessage: TypeAlias = Union[
    datetime,
    DecodedMessage,
    BaseModel,
    None,
]



from typing import Dict, List, Optional

from pydantic import BaseModel, Field

from propan.asyncapi.bindings import AsyncAPIOperationBinding
from propan.asyncapi.message import AsyncAPIMessage
from propan.asyncapi.utils import AsyncAPIExternalDocs, AsyncAPITag


class AsyncAPISubscription(BaseModel):
    operation_id: Optional[str] = Field(
        default=None,
        alias="operationId",
    )

    summary: Optional[str] = None
    description: Optional[str] = None

    bindings: Optional[AsyncAPIOperationBinding] = None

    message: AsyncAPIMessage

    security: Optional[Dict[str, List[str]]] = None

    # TODO
    # traits

    tags: Optional[List[AsyncAPITag]] = None
    external_docs: Optional[AsyncAPIExternalDocs] = Field(
        default=None,
        alias="externalDocs",
    )



from pathlib import Path
from typing import Sequence

from propan.cli.startproject.utils import write_file


def create_app_file(
    app_dir: Path,
    broker_class: str,
    imports: Sequence[str] = (),
    broker_init: Sequence[str] = ("    await broker.connect(settings.broker.url)",),
) -> None:
    write_file(app_dir / "__init__.py")

    write_file(
        app_dir / "serve.py",
        "import logging",
        "from typing import Optional",
        "",
        "import anyio",
        *imports,
        f"from propan import PropanApp, {broker_class}",
        f"from propan.annotations import {broker_class} as Broker, ContextRepo",
        "",
        "from config import init_settings",
        "from apps import router",
        "",
        "",
        f"broker = {broker_class}()",
        "broker.include_router(router)",
        "",
        "app = PropanApp(broker)",
        "",
        "",
        "@app.on_startup",
        "async def init_app(broker: Broker, context: ContextRepo, env: Optional[str] = None):",
        "    settings = init_settings(env)",
        '    context.set_global("settings", settings)',
        "",
        "    logger_level = logging.DEBUG if settings.debug else logging.INFO",
        "    app.logger.setLevel(logger_level)",
        "    broker.logger.setLevel(logger_level)",
        "",
        *broker_init,
        "",
        "",
        'if __name__ == "__main__":',
        "    anyio.run(app.run)",
    )


def create_handlers_file(
    filepath: Path,
    router_class: str,
) -> None:
    write_file(
        filepath,
        f"from propan import {router_class}",
        "from propan.annotations import Logger",
        "",
        f"router = {router_class}()",
        "",
        '@router.handle("test")',
        "async def base_handler(body: dict, logger: Logger):",
        "    logger.info(body)",
    )



import sys
from datetime import datetime
from types import MethodType
from typing import Any, Dict, Optional

from aiokafka.structs import ConsumerRecord

if sys.version_info < (3, 8):
    from asyncmock import AsyncMock
else:
    from unittest.mock import AsyncMock

from propan import KafkaBroker
from propan.test.utils import call_handler
from propan.types import SendableMessage

__all__ = (
    "build_message",
    "TestKafkaBroker",
)


def build_message(
    message: SendableMessage,
    topic: str,
    partition: Optional[int] = None,
    timestamp_ms: Optional[int] = None,
    key: Optional[bytes] = None,
    headers: Optional[Dict[str, str]] = None,
    *,
    reply_to: str = "",
) -> ConsumerRecord:
    msg, content_type = KafkaBroker._encode_message(message)
    k = key or b""
    headers = {
        "content-type": content_type or "",
        "reply_to": reply_to,
        **(headers or {}),
    }

    return ConsumerRecord(
        value=msg,
        topic=topic,
        partition=partition or 0,
        timestamp=timestamp_ms or int(datetime.now().timestamp()),
        timestamp_type=0,
        key=k,
        serialized_key_size=len(k),
        serialized_value_size=len(msg),
        checksum=sum(msg),
        offset=0,
        headers=[(i, j.encode()) for i, j in headers.items()],
    )


async def publish(
    self: KafkaBroker,
    message: SendableMessage,
    topic: str,
    key: Optional[bytes] = None,
    partition: Optional[int] = None,
    timestamp_ms: Optional[int] = None,
    headers: Optional[Dict[str, str]] = None,
    *,
    reply_to: str = "",
    callback: bool = False,
    callback_timeout: Optional[float] = None,
    raise_timeout: bool = False,
) -> Any:
    incoming = build_message(
        message=message,
        topic=topic,
        key=key,
        partition=partition,
        timestamp_ms=timestamp_ms,
        reply_to=reply_to,
        headers=headers,
    )

    for handler in self.handlers:  # pragma: no branch
        if topic in handler.topics:  # pragma: no branch
            r = await call_handler(
                handler, incoming, callback, callback_timeout, raise_timeout
            )
            if callback:  # pragma: no branch
                return r


def TestKafkaBroker(broker: KafkaBroker) -> KafkaBroker:
    broker.connect = AsyncMock()  # type: ignore
    broker.start = AsyncMock()  # type: ignore
    broker.publish = MethodType(publish, broker)  # type: ignore
    return broker



from typing import Awaitable, Callable, Optional, Sequence, Union

import aio_pika
from aio_pika.message import IncomingMessage
from fast_depends.dependencies import Depends
from typing_extensions import TypeAlias

from propan.brokers._model.broker_usecase import (
    AsyncDecoder,
    AsyncParser,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers._model.routing import BrokerRouter
from propan.brokers._model.schemas import PropanMessage
from propan.brokers.rabbit.schemas import RabbitExchange, RabbitQueue
from propan.types import AnyDict, SendableMessage

PikaSendableMessage: TypeAlias = Union[aio_pika.message.Message, SendableMessage]
RabbitMessage: TypeAlias = PropanMessage[IncomingMessage]

class RabbitRouter(BrokerRouter[IncomingMessage]):
    def handle(  # type: ignore[override]
        self,
        queue: Union[str, RabbitQueue],
        exchange: Union[str, RabbitExchange, None] = None,
        *,
        consume_arguments: Optional[AnyDict] = None,
        retry: Union[bool, int] = False,
        dependencies: Sequence[Depends] = (),
        decode_message: AsyncDecoder[IncomingMessage] = None,
        parse_message: AsyncParser[IncomingMessage] = None,
        # AsyncAPI
        description: str = "",
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[IncomingMessage, bool], Awaitable[T_HandlerReturn]],
    ]: ...



import logging
from enum import Enum
from ssl import SSLContext
from typing import Any, Awaitable, Callable, Dict, List, Optional, Sequence, Type, Union

import aio_pika
from aio_pika.message import IncomingMessage
from fastapi import params
from fastapi.datastructures import Default
from fastapi.routing import APIRoute
from fastapi.utils import generate_unique_id
from pamqp.common import FieldTable
from starlette import routing
from starlette.responses import JSONResponse, Response
from starlette.types import ASGIApp

from propan import RabbitBroker
from propan.brokers._model.broker_usecase import (
    AsyncDecoder,
    AsyncParser,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers.rabbit import RabbitExchange, RabbitQueue
from propan.fastapi.core import PropanRouter
from propan.log import access_logger
from propan.types import AnyDict

class RabbitRouter(PropanRouter[RabbitBroker]):
    def __init__(
        self,
        host: str = "localhost",
        port: int = 5672,
        login: str = "guest",
        password: str = "guest",
        virtualhost: str = "/",
        ssl: bool = False,
        ssl_options: Optional[aio_pika.abc.SSLOptions] = None,
        ssl_context: Optional[SSLContext] = None,
        timeout: aio_pika.abc.TimeoutType = None,
        client_properties: Optional[FieldTable] = None,
        # FastAPI kwargs
        prefix: str = "",
        tags: Optional[List[Union[str, Enum]]] = None,
        dependencies: Optional[Sequence[params.Depends]] = None,
        default_response_class: Type[Response] = Default(JSONResponse),
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        callbacks: Optional[List[routing.BaseRoute]] = None,
        routes: Optional[List[routing.BaseRoute]] = None,
        redirect_slashes: bool = True,
        default: Optional[ASGIApp] = None,
        dependency_overrides_provider: Optional[Any] = None,
        route_class: Type[APIRoute] = APIRoute,
        on_startup: Optional[Sequence[Callable[[], Any]]] = None,
        on_shutdown: Optional[Sequence[Callable[[], Any]]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        generate_unique_id_function: Callable[[APIRoute], str] = Default(
            generate_unique_id
        ),
        # Broker kwargs
        logger: Optional[logging.Logger] = access_logger,
        log_level: int = logging.INFO,
        log_fmt: Optional[str] = None,
        apply_types: bool = True,
        consumers: Optional[int] = None,
        decode_message: AsyncDecoder[IncomingMessage] = None,
        parse_message: AsyncParser[IncomingMessage] = None,
        schema_url: str = "/asyncapi",
        protocol: str = "amqp",
        protocol_version: str = "0.9.1",
    ) -> None:
        pass
    def add_api_mq_route(  # type: ignore[override]
        self,
        queue: Union[str, RabbitQueue],
        *,
        endpoint: HandlerCallable[T_HandlerReturn],
        exchange: Union[str, RabbitExchange, None] = None,
        retry: Union[bool, int] = False,
        decode_message: AsyncDecoder[IncomingMessage] = None,
        parse_message: AsyncParser[IncomingMessage] = None,
        description: str = "",
        dependencies: Optional[Sequence[params.Depends]] = None,
    ) -> Callable[[IncomingMessage, bool], Awaitable[T_HandlerReturn]]:
        pass
    def event(  # type: ignore[override]
        self,
        queue: Union[str, RabbitQueue],
        *,
        exchange: Union[str, RabbitExchange, None] = None,
        consume_arguments: Optional[AnyDict] = None,
        retry: Union[bool, int] = False,
        decode_message: AsyncDecoder[IncomingMessage] = None,
        parse_message: AsyncParser[IncomingMessage] = None,
        description: str = "",
        dependencies: Optional[Sequence[params.Depends]] = None,
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[IncomingMessage, bool], Awaitable[T_HandlerReturn]],
    ]:
        pass



import logging
import sys
from pathlib import Path
from typing import Dict, Optional

import anyio
import typer

from propan.__about__ import __version__
from propan.cli.docs import docs_app
from propan.cli.startproject import create_app
from propan.cli.utils.imports import get_app_path, try_import_propan
from propan.cli.utils.logs import LogLevels, get_log_level, set_log_level
from propan.cli.utils.parser import SettingField, parse_cli_args
from propan.log import logger

cli = typer.Typer(pretty_exceptions_short=True)
cli.add_typer(
    create_app, name="create", help="Create a new Propan project at [APPNAME] directory"
)
cli.add_typer(docs_app, name="docs", help="AsyncAPI scheme commands")


def version_callback(version: bool) -> None:
    if version is True:
        import platform

        typer.echo(
            "Running Propan %s with %s %s on %s"
            % (
                __version__,
                platform.python_implementation(),
                platform.python_version(),
                platform.system(),
            )
        )

        raise typer.Exit()


@cli.callback()
def main(
    version: Optional[bool] = typer.Option(
        False,
        "-v",
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show current platform, python and propan version",
    )
) -> None:
    """
    Generate, run and manage Propan apps to greater development experience
    """


@cli.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def run(
    ctx: typer.Context,
    app: str = typer.Argument(
        ..., help="[python_module:PropanApp] - path to your application"
    ),
    workers: int = typer.Option(
        1, show_default=False, help="Run [workers] applications with process spawning"
    ),
    log_level: LogLevels = typer.Option(
        LogLevels.info, case_sensitive=False, show_default=False, help="[INFO] default"
    ),
    reload: bool = typer.Option(
        False, "--reload", is_flag=True, help="Restart app at directory files changes"
    ),
) -> None:
    """Run [MODULE:APP] Propan application"""
    app, extra = parse_cli_args(app, *ctx.args)
    casted_log_level = get_log_level(log_level)

    module, app = get_app_path(app)

    app_dir = module.parent
    sys.path.insert(0, str(app_dir))

    args = (module, app, extra, casted_log_level)

    if reload and workers > 1:
        raise ValueError("You can't use reload option with multiprocessing")

    if reload is True:
        from propan.cli.supervisors.watchfiles import WatchReloader

        WatchReloader(target=_run, args=args, reload_dirs=(app_dir,)).run()

    elif workers > 1:
        from propan.cli.supervisors.multiprocess import Multiprocess

        Multiprocess(target=_run, args=(*args, logging.DEBUG), workers=workers).run()

    else:
        _run(module=module, app=app, extra_options=extra, log_level=casted_log_level)


def _run(
    module: Path,
    app: str,
    extra_options: Dict[str, SettingField],
    log_level: int = logging.INFO,
    app_level: int = logging.INFO,
) -> None:
    propan_app = try_import_propan(module, app)
    set_log_level(log_level, propan_app)

    propan_app._command_line_options = extra_options

    if sys.platform not in ("win32", "cygwin", "cli"):  # pragma: no cover
        try:
            import uvloop
        except ImportError:
            logger.warning("You have no installed `uvloop`")
        else:
            uvloop.install()

    anyio.run(propan_app.run, app_level)



from typing import Dict, Optional

from pydantic import BaseModel, Field, HttpUrl
from typing_extensions import Literal


class AsyncAPIOauthFlowObj(BaseModel):
    authorization_url: Optional[HttpUrl] = Field(
        default=None,
        alias="authorizationUrl",
    )
    token_url: Optional[HttpUrl] = Field(
        default=None,
        alias="tokenUrl",
    )
    refresh_url: Optional[HttpUrl] = Field(
        default=None,
        alias="refreshUrl",
    )
    scopes: Dict[str, str]


class AsyncAPIOauthFlows(BaseModel):
    implicit: Optional[AsyncAPIOauthFlowObj] = None
    password: Optional[AsyncAPIOauthFlowObj] = None
    client_credentials: Optional[AsyncAPIOauthFlowObj] = Field(
        default=None,
        alias="clientCredentials",
    )
    authorization_code: Optional[AsyncAPIOauthFlowObj] = Field(
        default=None,
        alias="authorizationCode",
    )


class AsyncAPISecuritySchemeComponent(BaseModel):
    type: Literal[
        "userPassword",
        "apikey",
        "X509",
        "symmetricEncryption",
        "asymmetricEncryption",
        "httpApiKey",
        "http",
        "oauth2",
        "openIdConnect",
        "plain",
        "scramSha256",
        "scramSha512",
        "gssapi",
    ]
    name: Optional[str] = None
    description: Optional[str] = None
    in_: Optional[str] = Field(
        default=None,
        alias="in",
    )
    scheme: Optional[str] = None
    bearer_format: Optional[str] = Field(
        default=None,
        alias="bearerFormat",
    )
    openid_connect_url: Optional[str] = Field(
        default=None,
        alias="openIdConnectUrl",
    )
    flows: Optional[AsyncAPIOauthFlows] = None



from typing import Any


class Singleton:
    _instance = None

    def __new__(cls, *args: Any, **kwargs: Any) -> "Singleton":
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    @classmethod
    def _drop(cls) -> None:
        cls._instance = None



from fast_depends import Depends
from fast_depends import inject as apply_types

from propan.utils.context import Context, ContextRepo, context
from propan.utils.no_cast import NoCast

__all__ = (
    "apply_types",
    "context",
    "Context",
    "ContextRepo",
    "Depends",
    "NoCast",
)



from aio_pika.message import IncomingMessage

from propan.brokers.rabbit import RabbitBroker
from propan.fastapi.core.router import PropanRouter


class RabbitRouter(PropanRouter[RabbitBroker, IncomingMessage]):
    broker_class = RabbitBroker



from propan import __about__ as about

try:
    from propan.fastapi.rabbit import RabbitRouter
except ImportError:
    RabbitRouter = about.INSTALL_RABBIT  # type: ignore

try:
    from propan.fastapi.redis import RedisRouter
except ImportError:
    RedisRouter = about.INSTALL_REDIS  # type: ignore

try:
    from propan.fastapi.kafka import KafkaRouter
except ImportError:
    KafkaRouter = about.INSTALL_KAFKA  # type: ignore

try:
    from propan.fastapi.nats import NatsJSRouter, NatsRouter
except ImportError:
    NatsRouter = NatsJSRouter = about.INSTALL_NATS  # type: ignore

try:
    from propan.fastapi.sqs import SQSRouter
except ImportError:
    SQSRouter = about.INSTALL_SQS  # type: ignore

assert any(
    (RabbitRouter, RedisRouter, KafkaRouter, NatsRouter, SQSRouter)
), about.INSTALL_MESSAGE

__all__ = (
    "RabbitRouter",
    "RedisRouter",
    "KafkaRouter",
    "NatsRouter",
    "SQSRouter",
    "NatsJSRouter",
)



import json
from contextlib import asynccontextmanager
from enum import Enum
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

from fastapi import APIRouter, FastAPI, Request, params
from fastapi.datastructures import Default
from fastapi.routing import APIRoute
from fastapi.utils import generate_unique_id
from starlette import routing
from starlette.responses import HTMLResponse, JSONResponse, Response
from starlette.routing import _DefaultLifespan
from starlette.types import AppType, ASGIApp, Lifespan
from typing_extensions import AsyncIterator, TypeVar

from propan.brokers._model.broker_usecase import (
    BrokerAsyncUsecase,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers._model.schemas import Queue
from propan.cli.docs.gen import (
    gen_app_schema_json,
    gen_app_schema_yaml,
    get_app_schema,
    json_schema_to_yaml,
    schema_to_json,
)
from propan.cli.docs.serving import get_asyncapi_html
from propan.fastapi.core.route import PropanRoute
from propan.types import AnyDict
from propan.utils.functions import to_async

Broker = TypeVar("Broker", bound=BrokerAsyncUsecase[Any, Any])
MsgType = TypeVar("MsgType")


class PropanRouter(APIRouter, Generic[Broker, MsgType]):
    broker_class: Type[Broker]
    broker: Broker
    docs_router: Optional[APIRouter]
    _after_startup_hooks: List[
        Callable[[AppType], Awaitable[Optional[Mapping[str, Any]]]]
    ]

    def __init__(
        self,
        *connection_args: Tuple[Any, ...],
        prefix: str = "",
        tags: Optional[List[Union[str, Enum]]] = None,
        dependencies: Optional[Sequence[params.Depends]] = None,
        default_response_class: Type[Response] = Default(JSONResponse),
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        callbacks: Optional[List[routing.BaseRoute]] = None,
        routes: Optional[List[routing.BaseRoute]] = None,
        redirect_slashes: bool = True,
        default: Optional[ASGIApp] = None,
        dependency_overrides_provider: Optional[Any] = None,
        route_class: Type[APIRoute] = APIRoute,
        on_startup: Optional[Sequence[Callable[[], Any]]] = None,
        on_shutdown: Optional[Sequence[Callable[[], Any]]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        schema_url: str = "/asyncapi",
        lifespan: Optional[Lifespan[Any]] = None,
        generate_unique_id_function: Callable[[APIRoute], str] = Default(
            generate_unique_id
        ),
        **connection_kwars: AnyDict,
    ) -> None:
        assert (
            self.broker_class
        ), "You should specify `broker_class` at your implementation"

        self.broker = self.broker_class(
            *connection_args,
            apply_types=False,
            **connection_kwars,  # type: ignore
        )

        super().__init__(
            prefix=prefix,
            tags=tags,
            dependencies=dependencies,
            default_response_class=default_response_class,
            responses=responses,
            callbacks=callbacks,
            routes=routes,
            redirect_slashes=redirect_slashes,
            default=default,
            dependency_overrides_provider=dependency_overrides_provider,
            route_class=route_class,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            generate_unique_id_function=generate_unique_id_function,
            lifespan=self.wrap_lifespan(lifespan),
            on_startup=on_startup,
            on_shutdown=on_shutdown,
        )

        if self.include_in_schema is True:  # pragma: no branch
            docs_router = APIRouter(
                prefix=prefix,
                tags=["asyncapi"],
                redirect_slashes=redirect_slashes,
                default=default,
                deprecated=deprecated,
            )
            docs_router.get(schema_url)(serve_asyncapi_schema)
            docs_router.get(f"{schema_url}.json")(download_app_json_schema)
            docs_router.get(f"{schema_url}.yaml")(download_app_yaml_schema)
            self.docs_router = docs_router
        else:
            self.docs_router = None

        self._after_startup_hooks = []

    def add_api_mq_route(
        self,
        path: Union[Queue, str],
        *extra: Union[Queue, str],
        endpoint: HandlerCallable[T_HandlerReturn],
        dependencies: Sequence[params.Depends],
        **broker_kwargs: AnyDict,
    ) -> Callable[[Any, bool], Awaitable[T_HandlerReturn]]:
        route = PropanRoute(
            path,
            *extra,
            endpoint=endpoint,
            dependencies=dependencies,
            dependency_overrides_provider=self.dependency_overrides_provider,
            broker=self.broker,
            **broker_kwargs,
        )
        self.routes.append(route)
        return route.handler

    def event(
        self,
        path: Union[str, Queue],
        *extra: Union[Queue, str],
        dependencies: Optional[Sequence[params.Depends]] = None,
        **broker_kwargs: Dict[str, Any],
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[MsgType, bool], Awaitable[T_HandlerReturn]],
    ]:
        current_dependencies = self.dependencies.copy()
        if dependencies:
            current_dependencies.extend(dependencies)

        def decorator(
            func: HandlerCallable[T_HandlerReturn],
        ) -> Callable[[MsgType, bool], Awaitable[T_HandlerReturn]]:
            return self.add_api_mq_route(
                path,
                *extra,
                endpoint=func,
                dependencies=current_dependencies,
                **broker_kwargs,
            )

        return decorator

    def wrap_lifespan(self, lifespan: Optional[Lifespan[Any]] = None) -> Lifespan[Any]:
        if lifespan is not None:
            lifespan_context = lifespan
        else:
            lifespan_context = _DefaultLifespan(self)

        @asynccontextmanager
        async def start_broker_lifespan(
            app: FastAPI,
        ) -> AsyncIterator[Mapping[str, Any]]:
            app.broker = self.broker  # type: ignore
            if self.docs_router:
                app.include_router(self.docs_router)

            async with lifespan_context(app) as maybe_context:
                if maybe_context is None:
                    context: Dict[str, Any] = {}
                else:
                    context = dict(maybe_context)

                context.update({"broker": self.broker})
                await self.broker.start()

                for h in self._after_startup_hooks:
                    h_context = await h(app)
                    if h_context:  # pragma: no branch
                        context.update(h_context)

                yield context
                await self.broker.close()

        return start_broker_lifespan

    def after_startup(
        self,
        func: Union[
            Callable[[AppType], Mapping[str, Any]],
            Callable[[AppType], Awaitable[Mapping[str, Any]]],
            Callable[[AppType], None],
            Callable[[AppType], Awaitable[None]],
        ],
    ) -> None:
        self._after_startup_hooks.append(to_async(func))  # type: ignore


def download_app_json_schema(r: Request) -> Response:
    return Response(
        content=json.dumps(
            gen_app_schema_json(r.app),
            indent=4,
        ),
        headers={
            "Content-Type": "application/octet-stream",
        },
    )


def download_app_yaml_schema(r: Request) -> Response:
    return Response(
        content=gen_app_schema_yaml(r.app),
        headers={
            "Content-Type": "application/octet-stream",
        },
    )


def serve_asyncapi_schema(
    r: Request,
    sidebar: bool = True,
    info: bool = True,
    servers: bool = True,
    operations: bool = True,
    messages: bool = True,
    schemas: bool = True,
    errors: bool = True,
    expandMessageExamples: bool = True,
) -> HTMLResponse:
    raw_schema = get_app_schema(r.app)
    json_schema = schema_to_json(raw_schema)
    schema = json_schema_to_yaml(json_schema)
    return HTMLResponse(
        content=get_asyncapi_html(
            schema,
            sidebar=sidebar,
            info=info,
            servers=servers,
            operations=operations,
            messages=messages,
            schemas=schemas,
            errors=errors,
            expand_message_examples=expandMessageExamples,
            title=raw_schema.info.title if raw_schema else "Propan",
        )
    )



from propan.brokers.sqs import SQSBroker
from propan.fastapi.core.router import PropanRouter
from propan.types import AnyDict


class SQSRouter(PropanRouter[SQSBroker, AnyDict]):
    broker_class = SQSBroker



class SkipMessage(Exception):
    """PushBackWatcher Instruction to skip message"""


WRONG_PUBLISH_ARGS = ValueError(
    "You should use `reply_to` to send response to long-living queue "
    "and `callback` to get response in sync mode."
)



import logging
from collections import defaultdict
from enum import Enum
from typing import DefaultDict, Optional, Union

from propan.cli.app import PropanApp


class LogLevels(str, Enum):
    critical = "critical"
    error = "error"
    warning = "warning"
    info = "info"
    debug = "debug"


LOG_LEVELS: DefaultDict[str, int] = defaultdict(
    lambda: logging.INFO,
    **{
        "critical": logging.CRITICAL,
        "error": logging.ERROR,
        "warning": logging.WARNING,
        "info": logging.INFO,
        "debug": logging.DEBUG,
    },
)


def get_log_level(level: Union[LogLevels, str, int]) -> int:
    if isinstance(level, int):
        return level

    if isinstance(level, LogLevels):
        return LOG_LEVELS[level.value]

    if isinstance(level, str):  # pragma: no branch
        return LOG_LEVELS[level.lower()]


def set_log_level(level: int, app: PropanApp) -> None:
    if app.logger:
        app.logger.setLevel(level)

    broker_logger: Optional[logging.Logger] = getattr(app.broker, "logger", None)
    if broker_logger is not None:
        broker_logger.setLevel(level)



from typing import Awaitable, Callable, Optional, Sequence, Union

from fast_depends.dependencies import Depends

from propan.brokers._model.broker_usecase import (
    AsyncDecoder,
    AsyncParser,
    HandlerCallable,
    T_HandlerReturn,
)
from propan.brokers._model.routing import BrokerRouter
from propan.brokers.sqs.schema import SQSQueue
from propan.types import AnyDict

class SQSRouter(BrokerRouter[AnyDict]):
    def handle(  # type: ignore[override]
        self,
        queue: Union[str, SQSQueue],
        *,
        wait_interval: int = 1,
        max_messages_number: int = 10,  # 1...10
        attributes: Sequence[str] = (),
        message_attributes: Sequence[str] = (),
        request_attempt_id: Optional[str] = None,
        visibility_timeout: int = 0,
        retry: Union[bool, int] = False,
        dependencies: Sequence[Depends] = (),
        decode_message: AsyncDecoder[AnyDict] = None,
        parse_message: AsyncParser[AnyDict] = None,
        description: str = "",
    ) -> Callable[
        [HandlerCallable[T_HandlerReturn]],
        Callable[[AnyDict, bool], Awaitable[T_HandlerReturn]],
    ]: ...



from nats.aio.msg import Msg

from propan.brokers.nats import NatsBroker, NatsJSBroker
from propan.fastapi.core.router import PropanRouter


class NatsRouter(PropanRouter[NatsBroker, Msg]):
    broker_class = NatsBroker


class NatsJSRouter(PropanRouter[NatsJSBroker, Msg]):
    broker_class = NatsJSBroker



from typing import Optional

from pydantic import BaseModel, Field

from propan.types import AnyDict


class AsyncAPISQSChannelBinding(BaseModel):
    queue: AnyDict
    version: str = Field(
        default="custom",
        alias="bindingVersion",
    )


class AsyncAPISQSOperationBinding(BaseModel):
    reply_to: Optional[AnyDict] = Field(default=None, alias="replyTo")

    version: str = Field(
        default="custom",
        alias="bindingVersion",
    )



from propan import PropanApp, RabbitBroker
from propan.annotations import Logger
from propan.brokers.rabbit import ExchangeType, RabbitExchange, RabbitQueue

broker = RabbitBroker()
app = PropanApp(broker)

topic_exchange = RabbitExchange("exchange", auto_delete=True, type=ExchangeType.TOPIC)
queue_1 = RabbitQueue("test-queue-1", auto_delete=True, routing_key="*.info")
queue_2 = RabbitQueue("test-queue-2", auto_delete=True, routing_key="*.debug")


@broker.handle(queue_1, topic_exchange)
async def base_handler1(logger: Logger):
    logger.info("base_handler1")


@broker.handle(queue_1, topic_exchange)
async def base_handler2(logger: Logger):
    logger.info("base_handler2")


@broker.handle(queue_2, topic_exchange)
async def base_handler3(logger: Logger):
    logger.info("base_handler3")


@app.after_startup
async def send_messages():
    await broker.publish(routing_key="logs.info", exchange=topic_exchange)
    await broker.publish(routing_key="logs.info", exchange=topic_exchange)
    await broker.publish(routing_key="logs.info", exchange=topic_exchange)
    await broker.publish(routing_key="logs.debug", exchange=topic_exchange)



import asyncio

from rocketry import Rocketry
from rocketry.args import Arg

from propan import RabbitBroker

app = Rocketry(execution="async")

broker = RabbitBroker()
app.params(broker=broker)


async def start_app():
    async with broker:
        await app.serve()


@app.task("every 1 second", execution="async")
async def publish(br: RabbitBroker = Arg("broker")):
    await br.publish("Hi, Rocketry!", "test")


if __name__ == "__main__":
    asyncio.run(start_app())
